import customtkinter as ctk
import os
import tkinter as tk
from tkinter import filedialog
from autogaita.gui.gui_utils import create_folder_icon

# This script contains functions to create widgets that are shared across GUIs.


def header_label(
    window,
    header_string,
    widget_cfg,
):
    """Header labels"""
    return ctk.CTkLabel(
        window,
        text=header_string,
        fg_color=widget_cfg["FG_COLOR"],
        text_color=widget_cfg["HEADER_TXT_COLOR"],
        font=(widget_cfg["HEADER_FONT_NAME"], widget_cfg["MAIN_HEADER_FONT_SIZE"]),
    )


def header_button(window, button_string, widget_cfg):
    """Header button"""
    return ctk.CTkButton(
        window,
        text=button_string,
        fg_color=widget_cfg["FG_COLOR"],
        hover_color=widget_cfg["HOVER_COLOR"],
        font=(widget_cfg["HEADER_FONT_NAME"], widget_cfg["HEADER_FONT_SIZE"]),
    )


def exit_button(window, widget_cfg):
    """Exit button"""
    return ctk.CTkButton(
        window,
        text="Exit",
        fg_color=widget_cfg["CLOSE_COLOR"],
        hover_color=widget_cfg["CLOSE_HOVER_COLOR"],
        font=(widget_cfg["HEADER_FONT_NAME"], widget_cfg["HEADER_FONT_SIZE"]),
    )


def label_and_entry_pair(
    window, label_text, entry_variable, widget_cfg, adv_cfg_textsize=False
):
    """Flexible label & entry widget-pair"""
    label = ctk.CTkLabel(
        window,
        text=label_text,
        font=(widget_cfg["TEXT_FONT_NAME"], widget_cfg["TEXT_FONT_SIZE"]),
    )
    entry = ctk.CTkEntry(
        window,
        textvariable=entry_variable,
        font=(widget_cfg["TEXT_FONT_NAME"], widget_cfg["TEXT_FONT_SIZE"]),
    )
    if adv_cfg_textsize:
        label.configure(
            font=(widget_cfg["TEXT_FONT_NAME"], widget_cfg["ADV_CFG_TEXT_FONT_SIZE"])
        )
        entry.configure(
            font=(widget_cfg["TEXT_FONT_NAME"], widget_cfg["ADV_CFG_TEXT_FONT_SIZE"])
        )
    return label, entry


def checkbox(window, label_text, entry_variable, widget_cfg, adv_cfg_textsize=False):
    """Checkbox"""
    checkbox = ctk.CTkCheckBox(
        window,
        text=label_text,
        variable=entry_variable,
        onvalue=True,
        offvalue=False,
        hover_color=widget_cfg["HOVER_COLOR"],
        fg_color=widget_cfg["FG_COLOR"],
        font=(widget_cfg["TEXT_FONT_NAME"], widget_cfg["TEXT_FONT_SIZE"]),
    )
    if adv_cfg_textsize:
        checkbox.configure(
            font=(widget_cfg["TEXT_FONT_NAME"], widget_cfg["ADV_CFG_TEXT_FONT_SIZE"])
        )
    return checkbox


def make_browse(
    parent_window,
    row,  # row of frame in parent_window
    column,  # column of frame in parent_window
    widget_cfg,
    entry_width=300,  # usual width of entry, needs to be less than 300 for group gui
    columnspan=None,
    sticky=None,
    text_var=None,  # for group gui where we dont have var_dict => only set text_var
    var_key=None,  # key for text_var in var_dict e.g. root_dir, results_dir
    var_dict=None,  # dictionary of var_key e.g. cfg, results
    is_file=False,  # ask for directory instead of file
    filetypes=None,  # to set default filetype
    initial_dir=None,  # function to get initial directory
    pady=None,
    adv_cfg_textsize=False,  # if true, use ADV_CFG_TEXT_FONT_SIZE
):
    """Entry & Browse button - placed in a frame. Function takes a lot of optional parameters to be able to deal with all the different use cases of this button across our different GUIs."""

    if (
        text_var is None and var_dict is not None
    ):  # we either need var_key AND var_dict or only text_var
        if (
            var_key not in var_dict
        ):  # make sure not to overwerite existing variables, important for configs
            var_dict[var_key] = tk.StringVar()
        text_var = var_dict[var_key]

    if is_file:
        filetypes = [("Excel files", "*.xlsx *.xls")]  # default filetype

    if adv_cfg_textsize:
        font_setting = (
            widget_cfg["TEXT_FONT_NAME"],
            widget_cfg["ADV_CFG_TEXT_FONT_SIZE"],
        )
    else:
        font_setting = (widget_cfg["TEXT_FONT_NAME"], widget_cfg["TEXT_FONT_SIZE"])

    # Frame
    frame = ctk.CTkFrame(parent_window, fg_color="transparent")
    frame.grid(row=row, column=column, sticky=sticky, columnspan=columnspan, pady=pady)

    # Entry
    entry = ctk.CTkEntry(
        frame, textvariable=text_var, font=font_setting, width=entry_width
    )
    entry.grid(row=0, column=1)

    def callback():  # function for browse buttons
        try:
            desktop_dir = os.path.join(os.path.expanduser("~"), "Desktop")
        except:
            desktop_dir = os.path.join(os.path.expanduser("~"))
        if initial_dir is None:
            dir_path = desktop_dir  # default directory: Desktop
        else:
            dir_path = initial_dir()  # call function to get initial directory
        if not os.path.exists(dir_path):
            dir_path = desktop_dir

        if is_file:
            selected = filedialog.askopenfilename(
                initialdir=dir_path, filetypes=filetypes
            )  # ask for file
        else:
            selected = filedialog.askdirectory(initialdir=dir_path)  # ask for directory
        if selected:
            entry.delete(0, tk.END)  # empty the entry first
            entry.insert(
                0, selected
            )  # insert the selected file or directory at index 0 of entry
            text_var.set(
                selected
            )  # set the chosen file as variable, e.g. "sctable_filename"
        parent_window.focus_force()  # bring focus back to root window or else button stays clicked

    # Browse button
    icon = create_folder_icon()  # load picture of folder icon for browse button
    button = ctk.CTkButton(
        frame,
        width=30,
        image=icon,
        text="",
        command=callback,
        fg_color=widget_cfg["FG_COLOR"],
        hover_color=widget_cfg["HOVER_COLOR"],
    )
    button.grid(row=0, column=2, padx=5)
    return entry



# ...................................  imports  ........................................
from autogaita.universal3D.universal3D_1_preparation import some_prep
from autogaita.universal3D.universal3D_2_sc_extraction import (
    extract_stepcycles,
    check_stepcycles,
)
from autogaita.universal3D.universal3D_3_analysis import analyse_and_export_stepcycles
from autogaita.universal3D.universal3D_4_plots import plot_results
from autogaita.resources.utils import print_finish, PlotPanel
import matplotlib
import matplotlib.pyplot as plt

# .................................  constants  ........................................
matplotlib.use("agg")
# Agg is a non-interactive backend for plotting that can only write to files
# this is used to generate and save the plot figures
# later a tkinter backend (FigureCanvasTkAgg) is used for the plot panel
plt.rcParams["figure.dpi"] = 300  # increase resolution of figures
from autogaita.gui.gui_constants import UNIVERSAL3D_FG_COLOR, UNIVERSAL3D_HOVER_COLOR

# %%
# ......................................................................................
# ........................  an important note for yourself  ............................
# ......................................................................................
# Please read this (& check_data_column_names!) when you are confused about col-names.
# It's a bit tricky in this code, see comments & doc about this issue in:
# 1. check_and_fix_cfg_strings (why am I here, read the others)
# 2. check_data_column_names (read me first)
# 3. add_features (then me)
# 4. plot_results (me if you really have to)
# => mainly (from 2.):
# Bodyside-specific colnames have to end WITHOUT a space (because we concat
# "name" + ", leg " + "Z" - so leg ends with a space)
# Bodyside-nonspecific colnames have to end WITH a space (because we concat "name "
# + "Z" so name has to end with a space)
# ......................................................................................


# .................................  main program  .....................................
def universal3D(info, folderinfo, cfg):
    """Runs the main program for a given subject's run

    Procedure
    ---------
    1) import & preparation
    2) step cycle extraction
    3) z-normalisation, y-flipping & feature computation for individual step cycles
    4) step cycle normalisaion, dataframe creation & XLS-exportation
    5) plots
    """
    # .............. initiate plot panel class and build loading screen ................
    # create class instance independently of "dont_show_plots" to not break the code
    plot_panel_instance = PlotPanel(UNIVERSAL3D_FG_COLOR, UNIVERSAL3D_HOVER_COLOR)

    if cfg["dont_show_plots"] is True:
        pass  # going on without building the loading screen

    elif cfg["dont_show_plots"] is False:  # -> show plot panel
        # build loading screen
        plot_panel_instance.build_plot_panel_loading_screen()

    # ...............................  preparation  ....................................
    data, global_Y_max = some_prep(info, folderinfo, cfg)
    if (data is None) & (global_Y_max is None):
        return

    # ..........................  step-cycle extraction  ...............................
    all_cycles = extract_stepcycles(data, info, folderinfo, cfg)
    all_cycles = check_stepcycles(all_cycles, info)
    if not all_cycles:  # only None if both leg's SCs were None
        if cfg["dont_show_plots"] is False:  # otherwise stuck at loading
            plot_panel_instance.destroy_plot_panel()
        return

    # ......  main analysis: y-flipping, features, df-creation & exports  ..............
    results = analyse_and_export_stepcycles(data, all_cycles, global_Y_max, info, cfg)

    # ..................................  plots  .......................................
    plot_results(results, all_cycles, info, cfg, plot_panel_instance)

    # ..............................  print finish  ....................................
    print_finish(info)


# ..................................  if we hit run  ...................................
if __name__ == "__main__":
    universal3D_info_message = (
        "\n*************\nnot like this\n*************\n"
        + "You are trying to execute autogaita.universal3D as a script, but that is not "
        + "possible.\nIf you prefer a non-GUI approach, please either: "
        + "\n1. Call this as a function, i.e. autogaita.universal3D(info, folderinfo, cfg)"
        + "\n2. Use the single or multirun scripts in the batchrun_scripts folder"
    )
    print(universal3D_info_message)



from autogaita.group.group_main import (
    group,
    import_data,
    avg_and_std,
    PCA_main,
    create_stats_df,
    cluster_extent_test,
    anova_design_sanity_check,
    tukeys_only_info_message,
    ANOVA_main,
    plot_results,
    print_finish,
)
from autogaita.group.group_1_preparation import some_prep
from autogaita.group.group_2_data_processing import (
    load_previous_runs_dataframes,
    check_PCA_and_stats_variables,
)
from autogaita.group.group_3_PCA import run_PCA, convert_PCA_bins_to_list
from autogaita.group.group_4_stats import run_ANOVA, multcompare_SC_Percentages
from autogaita.resources.utils import bin_num_to_percentages
import os
import math
import pytest
from sklearn import datasets
import pandas as pd
import pandas.testing as pdt
import numpy as np


# %%................................  fixtures  ........................................
@pytest.fixture
def extract_folderinfo(tmp_path):
    return {
        "group_names": ["group1", "group2"],
        "group_dirs": ["/path/to/group1", "/path/to/group2"],
        "results_dir": tmp_path,
        "load_dir": "",
    }


@pytest.fixture
def extract_cfg():
    return {
        "do_permtest": True,
        "do_anova": True,
        "permutation_number": 100,
        "PCA_n_components": 3,
        "PCA_custom_scatter_PCs": "",
        "PCA_save_3D_video": False,  # True
        "PCA_bins": "",
        "stats_threshold": 0.05,
        "plot_SE": False,
        "color_palette": "viridis",
        "dont_show_plots": True,
        "legend_outside": True,
        "which_leg": "left",
        "anova_design": "Mixed ANOVA",
        "PCA_variables": [],
        "stats_variables": [],
    }


# %%..............................  1. preparation  ....................................
def test_some_preps_cfg_files(extract_folderinfo, extract_cfg):
    """Check for equivalence of the cfg file using load dir and the cfg dict generated by some_prep"""
    # 1. Ensure that the cfg is equal to that used in group approval (with repo data)
    extract_folderinfo["group_names"] = ["5 mm", "12 mm", "25 mm"]
    extract_folderinfo["group_dirs"] = [
        "example data/5mm/Results/",
        "example data/12mm/Results/",
        "example data/25mm/Results/",
    ]
    extract_cfg["do_permtest"] = False
    extract_cfg["anova_design"] = "RM ANOVA"
    extract_cfg["PCA_variables"] = [
        "Knee y",
        "Ankle y",
        "Hind paw tao y",
        "Ankle Angle",
        "Knee Angle",
        "Hip Angle",
    ]
    extract_cfg["stats_variables"] = extract_cfg["PCA_variables"]
    # 2. Run some_prep twice with and without load dir & test cfg match
    extract_folderinfo, generated_cfg = some_prep(extract_folderinfo, extract_cfg)
    extract_folderinfo["load_dir"] = "example data/group"
    extract_folderinfo, loaded_cfg = some_prep(extract_folderinfo, extract_cfg)
    # 3. Check that cfgs are result of different processes (being loaded or not) and
    #    not exact copies (e.g. references to the same object) for some Python-reason
    assert "loaded" in loaded_cfg.keys()
    assert "loaded" not in generated_cfg.keys()
    # 4. Check that the cfgs are equivalent
    loaded_cfg.pop("loaded")  # except this key of course - get rid of it
    generated_cfg == loaded_cfg
    # 5. Finally change some cfg key and assert that it has not been overwritten by
    # loading a previous run's cfg
    # => might seem unnecessary but the way this was implemented initially did
    #    overwrite cfgs due to how it was loading stuff
    extract_cfg["do_permtest"] = True
    extract_cfg["anova_design"] = "Mixed ANOVA"
    extract_cfg["PCA_variables"] = ["Ankle Angle", "Knee Angle", "Hip Angle"]
    extract_folderinfo, loaded_cfg_2 = some_prep(extract_folderinfo, extract_cfg)
    differing_keys = ["do_permtest", "anova_design", "PCA_variables"]
    for key in differing_keys:
        assert loaded_cfg_2[key] != loaded_cfg[key]


# %%............................  2. data processing  ..................................
def test_check_PCA_and_stats_variables(extract_folderinfo, extract_cfg):
    """Test this sanity check does not have the necessary columns"""
    group_names = ["5 mm", "12 mm", "25 mm"]
    extract_folderinfo["group_names"] = group_names
    extract_folderinfo["load_dir"] = "example data/group"
    # changing PCA/stats vars like this already performs a first test because the load
    # function only runs successfully, if the test_PCA_.. function does
    # => i.e. the features are present in the all 3 kinds of dfs
    extract_cfg["PCA_variables"] = ["Ankle Angle", "Knee Angle"]
    extract_cfg["stats_variables"] = ["Ankle Angle", "Knee Angle"]
    avg_dfs, _, _, extract_cfg = load_repos_group_data(extract_folderinfo, extract_cfg)
    # test correct failure
    # => code means: test that after removing the column our check correctly raises a
    #    ValueError
    avg_dfs[0].drop(columns=["Ankle Angle"], inplace=True)
    with pytest.raises(ValueError):
        check_PCA_and_stats_variables(
            avg_dfs[0], "5 mm", "Average", extract_folderinfo, extract_cfg
        )


def test_smoke_load_dir(extract_folderinfo, extract_cfg):
    """Test that load dir runs through without errors and produces Results"""
    extract_folderinfo["group_names"] = ["5 mm", "12 mm", "25 mm"]
    extract_folderinfo["load_dir"] = "example data/group"
    group(extract_folderinfo, extract_cfg)
    # I checked and hardcoded the 39 here, shouldn't change.
    # If it does I want to know and this should fail
    assert len(os.listdir(extract_folderinfo["results_dir"])) == 39


def test_load_previous_runs_dataframes(extract_folderinfo, extract_cfg):
    """Testing if errors are raised correctly and if the loaded dfs are eqiuvalent to the ones import_data generates"""
    # NOTE
    # ----
    # => last tests fail on git actions but not locally
    # => quite sure it's because of how git actions' os lists the files when loading
    # => we have a similar issue in test_group_approval see os.getenv("CI") there
    # 1: fails as wanted if group name wrong (df not found in load dir)
    extract_folderinfo["group_names"] = ["not 5 mm", "12 mm", "25 mm"]
    extract_folderinfo["load_dir"] = "example data/group"
    with pytest.raises(FileNotFoundError):
        avg_dfs, _, _, extract_cfg = load_repos_group_data(
            extract_folderinfo, extract_cfg
        )
    # 2: avg_dfs equivalent to import_data's avg_dfs
    extract_folderinfo["group_names"] = ["5 mm", "12 mm", "25 mm"]
    extract_folderinfo["group_dirs"] = [  # for "import_data" later
        "example data/5mm/Results/",
        "example data/12mm/Results/",
        "example data/25mm/Results/",
    ]
    avg_dfs, _, _, extract_cfg = load_repos_group_data(extract_folderinfo, extract_cfg)
    # some prep required for import data & avg_and_std
    extract_cfg["sampling_rate"] = 100
    extract_cfg["bin_num"] = 25
    extract_cfg["save_to_xls"] = [True, True, True]
    extract_cfg["tracking_software"] = "DLC"
    extract_cfg["analyse_average_x"] = True
    i_dfs, _, extract_cfg = import_data(extract_folderinfo, extract_cfg)
    i_avg_dfs, _ = avg_and_std(i_dfs, extract_folderinfo, extract_cfg)
    if not os.getenv("CI"):
        for g in range(len(extract_folderinfo["group_names"])):
            pdt.assert_frame_equal(
                avg_dfs[g], i_avg_dfs[g], check_dtype=False, check_exact=False
            )
    # 3: check that it also works for Universal 3D
    # => only do it locally, I don't want to upload the MoVi Data to git
    if not os.getenv("CI"):
        extract_folderinfo["group_names"] = ["crawling", "jogging"]
        extract_folderinfo["group_dirs"] = [
            "/Users/mahan/sciebo/Research/AutoGaitA/Showcase 3/MOVI/Final/crawling/Results/",
            "/Users/mahan/sciebo/Research/AutoGaitA/Showcase 3/MOVI/Final/jogging/Results/",
        ]
        extract_folderinfo["load_dir"] = (
            "/Users/mahan/sciebo/Research/AutoGaitA/Showcase 3/MOVI/Unit Test of load_prev_groupruns_dfs/"
        )
        avg_dfs, _, _, extract_cfg = load_repos_group_data(
            extract_folderinfo, extract_cfg
        )
        extract_cfg["tracking_software"] = "Universal 3D"
        extract_cfg["which_leg"] = "right"
        extract_cfg["analyse_average_y"] = True
        i_dfs, _, extract_cfg = import_data(extract_folderinfo, extract_cfg)
        i_avg_dfs, _ = avg_and_std(i_dfs, extract_folderinfo, extract_cfg)
        if not os.getenv("CI"):
            for g in range(len(extract_folderinfo["group_names"])):
                pdt.assert_frame_equal(
                    avg_dfs[g], i_avg_dfs[g], check_dtype=False, check_exact=False
                )


# %%..............................  4. statistics  .....................................


def test_load_dfs_stats_df_perm_test_smoke(extract_folderinfo, extract_cfg):
    """Smoke test three functions:
    1. load_previous_runs-dataframes
    2. create_stats_df
    3. cluster_extent_test
    => Check if it all runs without errors on our repo's example data
    """
    # first, prepare approrpiately & overwrite the fixtures of this script as needed
    # 1) for load_previous_runs_dataframes (called by load repos group data)
    extract_folderinfo["group_names"] = ["5 mm", "12 mm", "25 mm"]
    extract_folderinfo["load_dir"] = "example data/group"
    avg_dfs, g_avg_dfs, g_std_dfs, extract_cfg = load_repos_group_data(
        extract_folderinfo, extract_cfg
    )
    # 2) for stats_df & the permutation test functions
    extract_cfg["sampling_rate"] = 100
    extract_cfg["do_permtest"] = True
    extract_cfg["dont_show_plots"] = True
    extract_cfg["tracking_software"] = "DLC"
    extract_cfg["group_color_dict"] = {"5 mm": "red", "12 mm": "blue", "25 mm": "green"}
    extract_folderinfo["contrasts"] = ["5 mm & 12 mm", "5 mm & 25 mm", "12 mm & 25 mm"]
    stats_df = create_stats_df(avg_dfs, extract_folderinfo, extract_cfg)
    plot_panel_instance = None
    cluster_extent_test(
        stats_df,
        g_avg_dfs,
        g_std_dfs,
        "Ankle Angle",
        extract_folderinfo,
        extract_cfg,
        plot_panel_instance,
    )


def test_Mixed_ANOVA(extract_cfg):
    # Adopted example data from https://real-statistics.com/
    # See the Mixed ANOVA xlsx file for complete link
    extract_cfg["do_anova"] = True
    extract_cfg["anova_design"] = "Mixed ANOVA"
    stats_df = pd.read_excel("tests/test_data/group_data/Mixed ANOVA Example Data.xlsx")
    stats_var = "Value"
    result = run_ANOVA(stats_df, stats_var, extract_cfg)
    # no GG p vals here
    assert math.isclose(result["p-unc"][0], stats_df["p(A)"][0], abs_tol=1e-05)
    assert math.isclose(result["p-unc"][1], stats_df["p(B)"][0], abs_tol=1e-05)
    assert math.isclose(result["p-unc"][2], stats_df["p(AxB)"][0], abs_tol=1e-05)


def test_multcomp_df_with_scipy_example(extract_folderinfo, extract_cfg):
    # Adopted example from https://docs.scipy.org/doc/scipy-1.15.0/reference/generated/
    # scipy.stats.tukey_hsd.html
    extract_folderinfo["group_names"] = ["group0", "group1", "group2"]
    extract_folderinfo["contrasts"] = [
        "group0 & group1",
        "group0 & group2",
        "group1 & group2",
    ]
    # fmt: off
    scipy_example = [24.5, 23.5, 26.4, 27.1, 29.9, 28.4, 34.2, 29.5, 32.2, 30.1, 26.1, 28.3, 24.3, 26.2, 27.8]  # fmt: on
    # prep data
    data = []
    bin_num = 25
    for i in scipy_example:
        data.append([i] * bin_num)
    data = np.ravel(data)
    # prep group_col
    subject_num = 5
    group_col = []
    for group in extract_folderinfo["group_names"]:
        for _ in range(subject_num):
            group_col.append([group] * bin_num)
    group_col = np.ravel(group_col)
    # prep ID_col
    total_subject_num = 15
    IDs = np.arange(1, total_subject_num + 1)
    ID_col = np.repeat(IDs, bin_num)
    # prep SC percentage col
    extract_cfg["bin_num"] = 25
    SC_percentages = bin_num_to_percentages(extract_cfg["bin_num"])
    SC_percentage_col = [SC_percentages] * total_subject_num
    SC_percentage_col = np.ravel(SC_percentage_col)
    # pytest.set_trace()
    # manually create dummy stats df
    stats_df = pd.DataFrame(data=data, columns=["Variable"])
    stats_df["Group"] = group_col
    stats_df["ID"] = ID_col
    stats_df["SC Percentage"] = SC_percentage_col
    # run the func and perform hard-coded check with vals of doc's example
    # => multcomp_df's rows are all equal because our SC percentages were all equal here
    # => correct results correspond to doc-example's rows 1, 4 and 2 first listing all
    #    stats-values, then p then CI low & CI high since that corresponds to the order
    #    of multcomp df's columns
    # => i.e., col1= SC %, then 3 cols for each contrast's q, then p, then CI low, then
    #    CI high in the order of groups0 & 1, 0 & 2, and 1 & 2
    multcomp_df = multcompare_SC_Percentages(
        stats_df, "Variable", extract_folderinfo, extract_cfg
    )
    # fmt: off
    correct_results = [-4.6, -0.260, 4.34, 0.014, 0.980, 0.02, -8.249, -3.909, 0.691, -0.951, 3.389, 7.989]  
    for r, result in enumerate(correct_results):
        assert math.isclose(result, multcomp_df.iloc[0, r+1], abs_tol=0.001)


# %%..................................  PCA  ...........................................
def test_run_PCA(extract_cfg):
    # Replicate the example found in https://www.kdnuggets.com/2023/05/
    # principal-component-analysis-pca-scikitlearn.html using our PCA df and PCA_info
    # structure
    # fmt: off
    true_results_PCA_eigenvectors = [[0.1443294, -0.24518758, -0.00205106, -0.23932041,  0.14199204,  0.39466085, 0.4229343, -0.2985331, 0.31342949, -0.0886167,0.29671456,  0.37616741, 0.28675223],
    [-0.48365155, -0.22493093, -0.31606881,  0.0105905,  -0.299634,   -0.06503951
    , 0.00335981, -0.02877949, -0.03930172, -0.52999567,  0.27923515,  0.16449619,
    -0.36490283],
    [-0.20738262,  0.08901289,  0.6262239,   0.61208035,  0.13075693,  0.14617896,
    0.1506819,   0.17036816,  0.14945431, -0.13730621,  0.08522192,  0.16600459,
    -0.12674592]]
    true_results_PCA_explained_var = [0.36198848, 0.1920749 , 0.11123631]
    # fmt: on
    wine_data = datasets.load_wine(as_frame=True)
    wine_df = wine_data.data
    features = wine_df.columns
    PCA_df, PCA_info = run_PCA(wine_df, features, extract_cfg)
    for i in range(3):
        # absolute values are compared because the signs can be different w. eigenvecs
        assert np.allclose(
            np.absolute(PCA_info["eigenvectors"][i]),
            np.absolute(true_results_PCA_eigenvectors[i]),
            atol=1e-05,
        )
    assert np.allclose(
        PCA_info["explained_vars"], true_results_PCA_explained_var, atol=1e-05
    )


cases = (
    (
        "1-30,50,70-100",
        100,
        np.arange(1, 31).tolist() + [50] + np.arange(70, 101).tolist(),
    ),
    ("1-70", 25, [4, 8, 12, 16, 20, 24, 28, 32, 36, 40, 44, 48, 52, 56, 60, 64, 68]),
    ("2, 10, 22, 44, 50, 79", 50, [2, 10, 22, 44, 50]),
    ("1-50,70-80", 10, [10, 20, 30, 40, 50, 70, 80]),
)  # fmt: skip
@pytest.mark.parametrize("PCA_bins, bin_num, expected_bins_list", cases)
def test_convert_PCA_bins_to_list(
    PCA_bins, bin_num, expected_bins_list, extract_folderinfo, extract_cfg
):
    extract_cfg["PCA_bins"] = PCA_bins
    extract_cfg["bin_num"] = bin_num
    updated_cfg = convert_PCA_bins_to_list(extract_folderinfo, extract_cfg)
    assert np.array_equal(updated_cfg["PCA_bins"], np.array(expected_bins_list))


def load_repos_group_data(extract_folderinfo, extract_cfg):
    """Use load_previous_runs_dataframes to load example data from the repo"""
    extract_cfg["sampling_rate"] = 100
    extract_cfg["group_color_dict"] = {"5 mm": "red", "12 mm": "blue", "25 mm": "green"}
    extract_folderinfo["contrasts"] = ["5 mm & 12 mm", "5 mm & 25 mm", "12 mm & 25 mm"]
    avg_dfs, g_avg_dfs, g_std_dfs, extract_cfg = load_previous_runs_dataframes(
        extract_folderinfo, extract_cfg
    )
    return avg_dfs, g_avg_dfs, g_std_dfs, extract_cfg


# NOTE !!!
# => This test is outdated and wrong because it is for two within subject factors (for
#    example 3 separate behavioural tests and pre/post medication treatment or so -
#    done by each subject)
# => I will not delete this since I might want to re-use it if I should support such
#    designs
#
# def test_twoway_RM_ANOVA(extract_cfg):
#     # Adopted example data from https://real-statistics.com/
#     # See the RM ANOVA xlsx file for complete link
#     extract_cfg["do_anova"] = True
#     extract_cfg["anova_design"] = "RM ANOVA"
#     stats_df = pd.read_excel("tests/test_data/group_data/RM ANOVA Example Data.xlsx")
#     stats_var = "Value"
#     result = run_ANOVA(stats_df, stats_var, extract_cfg)
#     # Note that the last 2 assert statements have a different tolerance because those
#     # p-values differed a bit. Not sure why but it's a tiny amount and tolerable IMO
#     pytest.set_trace()
#     assert math.isclose(result["p-unc"][0], stats_df["p(A)"][0], abs_tol=1e-05)
#     assert math.isclose(result["p-unc"][1], stats_df["p(B)"][0], abs_tol=1e-05)
#     assert math.isclose(result["p-unc"][2], stats_df["p(AxB)"][0], abs_tol=1e-05)
#     assert math.isclose(result["p-GG-corr"][0], stats_df["GG-p(A)"][0], abs_tol=1e-05)
#     assert math.isclose(result["p-GG-corr"][1], stats_df["GG-p(B)"][0], abs_tol=1e-04)
#     assert math.isclose(result["p-GG-corr"][2], stats_df["GG-p(AxB)"][0], abs_tol=1e-02)



def transform_joint_and_leg_to_colname(joint, legname, feature):
    """For Human Data: Transform a joint and leg name to Universal 3D-column name"""
    return joint + ", " + legname + " " + feature


def extract_feature_column(df, joint, legname, feature):
    """Extract the column of a given joint (or angle) x legname (or not) x feature combo

    Note
    ----
    Only use this when using .iloc, not .loc!
    ==> the return statement gives the column-index to be used with .iloc!
    """
    if joint + feature in df.columns:
        string = joint + feature
    else:
        string = transform_joint_and_leg_to_colname(joint, legname, feature)
    return df.columns.get_loc(string)



# %% imports
from autogaita.resources.utils import write_issues_to_textfile
import os
import json
import matplotlib.pyplot as plt
import seaborn as sns

# %% constants
from autogaita.resources.constants import ISSUES_TXT_FILENAME, CONFIG_JSON_FILENAME
from autogaita.group.group_constants import (
    GROUP_CONFIG_TXT_FILENAME,
    STATS_TXT_FILENAME,
    MULTCOMP_EXCEL_FILENAME_1,
    MULTCOMP_EXCEL_FILENAME_2,
    ORIG_SHEET_NAME,
    CONTRAST_SPLIT_STR,
)


# %% .................  workflow step #1 - unpack & prepare vars  ....................


def some_prep(folderinfo, cfg):
    """Add some folderinfo & cfg variables to the dictionaries for further processes"""

    # AN IMPORTANT NOTE ABOUT LOAD_DIR
    # --------------------------------
    # Alright so the group pipeline's cfg (and thus, of course, config.json) is a bit
    # special because it includes:
    # 1) first-level config-keys, such as "joints" or "angles", that reflect what has
    #    been analysed at the first level. Some of these are used in group workflow
    #    (e.g. tracking_software when plotting or sampling_rate for PCA) and all of
    #    these are  checked for equivalence across groups when running this without
    #    load_dir (see the for g_idx loop below) to ensure we are not comparing
    #    different sampling rates or so with a group analysis
    # 2) group-level config-keys, such as "do_permtest" or "PCA_variables" that define
    #    how group analysis should be done
    # Now:
    # When loading previously generated group dfs (i.e., using load_dir), the vars in
    # (2) should naturally be changing so the user can change "PCA_variables" or
    # "do_anova". The config.json is coded to reflect the group-keys of the most
    # recent analysis. The "first-level" config keys are, however, just checked for
    # equivalence once and then never changed by group gaita. So if users should
    # repeatedly run analyses in the same results_dir, the config.json file includes
    # the first-level keys of the first run and the group-level keys of the most recent
    # run. This is not an issue per se but very likely something that I might forget in
    # a year thus here is a note.

    # unpack
    group_names = folderinfo["group_names"]
    group_dirs = folderinfo["group_dirs"]
    results_dir = folderinfo["results_dir"]

    # names & paths to folderinfo
    if not os.path.isdir(results_dir):
        os.makedirs(results_dir)
    folderinfo["contrasts"] = []
    for i in range(len(group_names)):
        for j in range(i + 1, len(group_names)):
            contrast = group_names[i] + CONTRAST_SPLIT_STR + group_names[j]
            folderinfo["contrasts"].append(contrast)

    # check if we previously had saved info files, if so delete them
    for info_file_name in [
        GROUP_CONFIG_TXT_FILENAME,
        ISSUES_TXT_FILENAME,
        STATS_TXT_FILENAME,
        MULTCOMP_EXCEL_FILENAME_1,
        MULTCOMP_EXCEL_FILENAME_2,
    ]:
        info_file_path = os.path.join(results_dir, info_file_name)
        if os.path.exists(info_file_path):
            os.remove(info_file_path)

    # load a couple necessary first-level cfg vars from previous run's group config.json
    if len(folderinfo["load_dir"]) > 0:
        cfg = load_previous_runs_first_level_cfg_vars(folderinfo, cfg)

    # define save_to_xls and test PCA
    cfg = extract_save_to_xls_and_test_PCA_config(folderinfo, cfg)

    # if not loading previous results, ensure cfg-keys are equivalent across groups
    # then add them to cfg dict
    if len(folderinfo["load_dir"]) == 0:
        for g_idx, group_dir in enumerate(group_dirs):
            with open(
                os.path.join(group_dir, CONFIG_JSON_FILENAME), "r"
            ) as config_json_file:
                config_vars_from_json = json.load(config_json_file)
                for key in config_vars_from_json.keys():
                    # assigning like this ensure all keys are in all jsons across groups
                    if g_idx == 0:
                        cfg[key] = config_vars_from_json[key]
                    else:
                        # sanity check for group-differences in cfg variables!
                        if (key not in cfg.keys()) | (
                            cfg[key] != config_vars_from_json[key]
                        ):
                            error_message = (
                                "config.json variables differ between groups!"
                                + "\nPlease make sure that all cfg variables between "
                                + "groups match & try again!"
                                + f"\nMismatch at {key} in group {group_names[g_idx]}"
                            )
                            raise ValueError(error_message)
                        else:
                            cfg[key] = config_vars_from_json[key]

    # rename hind_joints to joints (if DLC or SLEAP)
    if "hind_joints" in cfg.keys():
        cfg["joints"] = cfg["hind_joints"]

    # update cfg keys in json file in results_dir
    # => i.e. if there's already a config.json @ results_dir (happens if load_dir was
    #    True) the if-condition below updates (only) the group-config keys according to
    #    this run's cfg dict
    # => this also means that first-level cfg keys are never changed (which is intended)
    config_json_path = os.path.join(results_dir, CONFIG_JSON_FILENAME)
    if os.path.exists(config_json_path):
        with open(config_json_path, "r") as config_json_file:
            existing_cfg = json.load(config_json_file)
        existing_cfg.update({key: cfg[key] for key in cfg if key in existing_cfg})
        cfg = existing_cfg  # update cfg with existing keys
    with open(config_json_path, "w") as config_json_file:
        json.dump(cfg, config_json_file)

    # create this plot stuff manually (cycler objects cannot be written to json)
    cfg["group_color_cycler"] = plt.cycler(
        "color", sns.color_palette(cfg["color_palette"], len(group_names))
    )
    cfg["group_color_dict"] = dict(
        zip(group_names, cfg["group_color_cycler"].by_key()["color"])
    )
    cfg["joint_color_cycler"] = plt.cycler(
        "color", sns.color_palette(cfg["color_palette"], len(cfg["joints"]))
    )
    cfg["angle_color_cycler"] = plt.cycler(
        "color", sns.color_palette(cfg["color_palette"], len(cfg["angles"]["name"]))
    )

    # have this key for a unit test - make sure it's never written to json
    if len(folderinfo["load_dir"]) > 0:
        cfg["loaded"] = True

    return folderinfo, cfg


def load_previous_runs_first_level_cfg_vars(folderinfo, cfg):
    """There are a couple "first-level" cfg vars (like "joints") we require for group gaita's workflow - load them here"""
    with open(
        os.path.join(folderinfo["load_dir"], CONFIG_JSON_FILENAME), "r"
    ) as config_json_file:
        old_cfg = json.load(config_json_file)
        cfg["sampling_rate"] = old_cfg["sampling_rate"]
        cfg["save_to_xls"] = old_cfg["save_to_xls"]
        cfg["joints"] = old_cfg["joints"]
        cfg["angles"] = old_cfg["angles"]
        cfg["tracking_software"] = old_cfg["tracking_software"]
        if "analyse_average_x" in old_cfg.keys():
            cfg["analyse_average_x"] = old_cfg["analyse_average_x"]
        if "analyse_average_y" in old_cfg.keys():
            cfg["analyse_average_y"] = old_cfg["analyse_average_y"]
    return cfg


def extract_save_to_xls_and_test_PCA_config(folderinfo, cfg):
    """Extract save_to_xls from example Normalised dfs and sanity check
    that they match between groups. Also some tests for users' PCA config!
    """

    # NOTE
    # ----
    # save_to_xls is a list of bools that is infered from file type of group's sheet
    # files - only when not using load_dir. if we use load_dir, save_to_xls is loaded
    # by load_previous_runs_first_level_cfg_vars

    # ................................  save_to_xls  ...................................
    if len(folderinfo["load_dir"]) == 0:
        # infer save_to_xls from sheet files
        cfg["save_to_xls"] = infer_save_to_xls_from_group_dirs_sheetfiles(folderinfo)

    # .........................  test if PCA config is valid  ..........................
    # only test if user wants PCA (ie. selected any features) and is not using the
    # var-explained appproiach
    if cfg["PCA_variables"] and cfg["PCA_n_components"] > 1:
        if len(cfg["PCA_variables"]) < cfg["PCA_n_components"]:
            PCA_variable_num = len(cfg["PCA_variables"])
            PCA_variables_str = "\n".join(cfg["PCA_variables"])
            PCA_error_message = (
                "\n*********\n! ERROR !\n*********\n"
                + "\nPCA Configuration invalid, number of input features must exceed "
                + "the number of principal components to compute!\n"
                + str(PCA_variable_num)
                + " PCA variables: \n"
                + PCA_variables_str
                + "\n & Number of wanted PCs: "
                + str(cfg["PCA_n_components"])
                + "\n Fix & re-run!"
            )
            write_issues_to_textfile(PCA_error_message, folderinfo)
            raise ValueError(PCA_error_message)
        if cfg["PCA_n_components"] < 2:
            print(
                "\n***********\n! WARNING !\n***********\n"
                + "Number of principal components of PCA cannot be 0 or 1!"
                + "\nRunning PCA on 2 components - if you do not want to perform PCA, "
                + "just don't choose any variables for it."
            )
            cfg["PCA_n_components"] = 2  # make sure to update in cfg dict
    # small fix to ensure that PCA vars don't have duplicates (this would have severe
    # consequences for run_PCA since features of PCA_model output and "my" features var
    # would not match)
    # => it should never happen via GUI really but I did this myself using group_dlcrun #    while not paying attention...
    unique_PCA_vars = list(set(cfg["PCA_variables"]))
    if len(unique_PCA_vars) != len(cfg["PCA_variables"]):
        duplicate_vars = [
            var for var in cfg["PCA_variables"] if cfg["PCA_variables"].count(var) > 1
        ]
        PCA_duplicates_error_message = (
            "\n*********\n! WARNING !\n*********\n"
            + "\nWe found duplicates in your PCA variables list!"
            + "\nWe removed them for you!"
            + "\n\nDuplicate variables were:\n"
            + "\n".join(list(set(duplicate_vars)))
        )
        print(PCA_duplicates_error_message)
        write_issues_to_textfile(PCA_duplicates_error_message, folderinfo)
        cfg["PCA_variables"] = unique_PCA_vars
    # check if PCA bin num is valid, two tests and 1 fix
    # => note there is separate code that transforms this string input into an int-list
    if cfg["PCA_bins"]:
        PCA_bins_error_message = ""
        cfg["PCA_bins"] = cfg["PCA_bins"].replace(" ", "")  # remove spaces
        # test 1: only digits, commas and dashes allowed
        for char in cfg["PCA_bins"]:
            if char.isdigit() or char in ["-", ","]:
                pass
            else:
                PCA_bins_error_message = (
                    "\n*********\n! ERROR !\n*********\n"
                    + "\nPCA Bin Number invalid!"
                    + "\nPlease use only numbers, commas and dashes in your input!"
                    + "\nFix & re-run!"
                )
        # test 2: no comma next to dash
        if ",-" in cfg["PCA_bins"] or "-," in cfg["PCA_bins"]:
            PCA_bins_error_message = (
                "\n*********\n! ERROR !\n*********\n"
                + "\nPCA Bin Number invalid!"
                + "\nCommas cannot be next to dashes in your input!"
                + "\nFix & re-run!"
            )
        # if any test failed we raise an error
        # => note we could continue execution without custom PCA bins but if the user
        #    tried to set them, it's likely important to them to have custom bins
        #    so we just stop everything and help them fix it
        if PCA_bins_error_message:
            print(PCA_bins_error_message)
            write_issues_to_textfile(PCA_bins_error_message, folderinfo)
            raise ValueError(PCA_bins_error_message)
        # fix for users: remove first and last characters if not digits
        while not cfg["PCA_bins"][0].isdigit():
            cfg["PCA_bins"] = cfg["PCA_bins"][1:]
        while not cfg["PCA_bins"][-1].isdigit():
            cfg["PCA_bins"] = cfg["PCA_bins"][:-1]

    return cfg


def infer_save_to_xls_from_group_dirs_sheetfiles(folderinfo):
    """Generate a list of save_to_xls bools that is automatically inferred from sheet file in group dir"""

    # unpack
    group_names = folderinfo["group_names"]
    group_dirs = folderinfo["group_dirs"]

    save_to_xls = [None] * len(group_dirs)
    for g, group_dir in enumerate(group_dirs):
        all_results_folders = os.listdir(
            group_dir
        )  # remove no-results valid_results_folders
        valid_results_folders = []
        # => Note if there are mixed filetypes, we set save_to_xls to True!
        sheet_type_mismatch_message = (
            "\n***********\n! WARNING !\n***********\n"
            + "Mismatch in sheet file types for group "
            + group_names[g]
            + "!\nSaving all output sheets to"
            + ".xlsx!\nRe-run first level & only save .csvs if "
            + "you want .csv files of group results!"
        )
        for folder in all_results_folders:
            # create save_to_xls here, there are two cases we have to deal with:
            # case 1: we found a csv file
            if os.path.exists(
                os.path.join(
                    group_dir,
                    folder,
                    folder + " - " + ORIG_SHEET_NAME + ".csv",
                )
            ):
                valid_results_folders.append(folder)
                if save_to_xls[g] is None:
                    save_to_xls[g] = False
                if save_to_xls[g] is True:
                    print(sheet_type_mismatch_message)
                    write_issues_to_textfile(sheet_type_mismatch_message, folderinfo)
            # case 2: we found a xlsx file
            elif os.path.exists(
                os.path.join(
                    group_dir,
                    folder,
                    folder + " - " + ORIG_SHEET_NAME + ".xlsx",
                )
            ):
                valid_results_folders.append(folder)
                if save_to_xls[g] is None:
                    save_to_xls[g] = True
                if save_to_xls[g] is False:
                    save_to_xls[g] = True
                    print(sheet_type_mismatch_message)
                    write_issues_to_textfile(sheet_type_mismatch_message, folderinfo)
        # test that at least 1 folder has valid results for all groups
        if not valid_results_folders:
            no_valid_results_error = (
                "\n*********\n! ERROR !\n*********\n"
                + "No valid results folder found for "
                + group_names[g]
                + "\nFix & re-run!"
            )
            print(no_valid_results_error)
            write_issues_to_textfile(no_valid_results_error, folderinfo)
    return save_to_xls



# %% imports
from autogaita.resources.utils import write_issues_to_textfile
import os
import pandas as pd
import numpy as np

# %% constants
from autogaita.universal3D.universal3D_constants import (
    TIME_COL,
    LEGS,
    SCXLS_SUBJCOLS,
    SCXLS_LEGCOLS,
    SCXLS_RUNCOLS,
    SCXLS_SCCOLS,
    SWINGSTART_COL,
    STANCEEND_COL,
)


# %% workflow step #2 - SC extraction (reading user-provided SC Table)


# ...............................  outer function  .....................................
def extract_stepcycles(data, info, folderinfo, cfg):
    """Read XLS file with SC annotations, find correct row & return all_cycles"""
    # unpack
    root_dir = folderinfo["root_dir"]
    sctable_filename = folderinfo["sctable_filename"]

    # load the table - try some filename & ending options
    if os.path.exists(os.path.join(root_dir, sctable_filename)):
        SCdf_full_filename = os.path.join(root_dir, sctable_filename)
    elif os.path.exists(os.path.join(root_dir, sctable_filename) + ".xlsx"):
        SCdf_full_filename = os.path.join(root_dir, sctable_filename) + ".xlsx"
    elif os.path.exists(os.path.join(root_dir, sctable_filename) + ".xls"):
        SCdf_full_filename = os.path.join(root_dir, sctable_filename) + ".xls"
    else:
        no_sc_table_message = (
            "No Annotation Table found! sctable_filename has to be @ root_dir"
        )
        raise FileNotFoundError(no_sc_table_message)
    # check if we need to specify engine (required for xlsx)
    try:
        SCdf = pd.read_excel(SCdf_full_filename)
    except:
        SCdf = pd.read_excel(SCdf_full_filename, engine="openpyxl")

    # extract & return all_cycles
    all_cycles = {"left": [], "right": []}
    for legname in LEGS:
        all_cycles[legname] = read_SC_info(data, SCdf, info, legname, cfg)
    return all_cycles


# ...........................  inner (main) function  ..................................
def read_SC_info(data, SCdf, info, legname, cfg):
    """Read table, and create a list of start/end indices of a leg's SCs"""
    # ...............................  preparation  ....................................
    # unpack
    name = info["name"]
    sampling_rate = cfg["sampling_rate"]

    # very first sanity check - see if table columns are labelled correctly
    valid_col_flags = [
        False,
        False,
        False,
        False,
    ]  # for user typos
    header_columns = ["", "", "", ""]
    for h, header in enumerate(
        [SCXLS_SUBJCOLS, SCXLS_LEGCOLS, SCXLS_RUNCOLS, SCXLS_SCCOLS]
    ):
        for header_col in header:
            if header_col in SCdf.columns:
                valid_col_flags[h] = True
                header_columns[h] = header_col
                break
    if not all(valid_col_flags):
        this_message = (
            "\n******************\n! CRITICAL ERROR !"
            + "\n******************\n"
            + "Annotation Table Column names are wrong!\n"
            + "Check Instructions!"
        )
        print(this_message)
        write_issues_to_textfile(this_message, info)
        return
    # some prep
    leg_col = SCdf.columns.get_loc(header_columns[1])  # INDEXING! (see list above)
    run_col = SCdf.columns.get_loc(header_columns[2])
    sc_col = SCdf.columns.get_loc(header_columns[3])
    # first find the rows of this leg
    # a. find overall start row of this subject
    SCdf[header_columns[0]] = SCdf[header_columns[0]].astype(str)  # ensure no ints!
    start_row = SCdf.index[SCdf[header_columns[0]] == name]
    if start_row.empty:
        this_message = (
            "\n******************\n! CRITICAL ERROR !\n******************\n"
            + "\nNo timestamp information found for ID: "
            + name
            + "\nCheck your Annotation Table & try again!"
        )
        print(this_message)
        write_issues_to_textfile(this_message, info)
        return
    # b. check if this subjects ID was found more than once, if so - stop!
    if len(start_row) > 1:
        this_message = (
            "\n******************\n! CRITICAL ERROR !\n******************\n"
            + "\nID "
            + name
            + " was found more than once in ID column!"
            + "\nCheck your Annotation Table & try again!"
        )
        print(this_message)
        write_issues_to_textfile(this_message, info)
        return
    # c. find start row of current leg (whileloop works for both legs)
    while SCdf.iloc[start_row, leg_col].values[0] != legname:
        start_row += 1
    end_row = start_row
    while isinstance(SCdf.iloc[end_row, run_col].values[0], int) & (
        end_row != len(SCdf) - 1
    ):
        end_row += 1
    # quick sanity checks that user input was correct
    if end_row != start_row:
        if not np.isnan(SCdf.iloc[end_row, leg_col].values[0]):
            this_message = (
                "\n******************\n! CRITICAL ERROR !\n******************\n"
                + "\nID: "
                + name
                + ", Leg: "
                + legname
                + "\nRun & Leg columns of Annotation "
                + "Table seem wrong! \nYou need to have an "
                + "empty row before each new leg or subject!"
                + "\nCheck your table & make sure that it "
                + "matches our template."
            )
            print(this_message)
            write_issues_to_textfile(this_message, info)
            return  # return None for this leg if formatting bad

    # ..............................  main xls read  ...................................
    # extract all runs of this subject / leg combination
    # ==> note that slicing behaves differently based on whether end_row is
    #     the last row of the dataframe or not. handle this here. Particularly:
    # ==> end_row is 1) first row of new subject or 2) SC end. If 1) we want
    #     end_row to not be included, if 2) we want it to be included
    # case 1) - end_row is nan row after last SC. iloc slicing is exclusive for endidx!
    if end_row[0] != (len(SCdf) - 1):
        runs = SCdf.iloc[int(start_row[0]) : int(end_row[0]), run_col]
    # case 2) - end_row is last row df. iloc with [startidx:, col] = inclusive endrow!
    else:
        runs = SCdf.iloc[int(start_row[0]) :, run_col]

    # find out the total number of scs & see if it matches user-provided values
    # (handle same two cases as above for runs with if/else)
    if end_row[0] != (len(SCdf) - 1):
        user_scnum = sum(SCdf.iloc[int(start_row[0]) : int(end_row[0]), sc_col])
    else:
        user_scnum = sum(SCdf.iloc[int(start_row[0]) :, sc_col])
    total_scnum = 0  # for sanity check (before warning-message below)
    run_scnums = [None for s in range(len(runs))]
    for r, run in enumerate(runs):
        run_row = runs[runs == run].index
        run_scnums[r] = 0
        for column in SCdf.columns:
            if STANCEEND_COL in column:
                if not np.isnan(SCdf[column][run_row].values[0]):
                    total_scnum += 1
                    run_scnums[r] += 1
    if user_scnum != total_scnum:  # warn the user, take the values we found
        this_message = (
            "\n***********\n! WARNING !\n***********\n"
            + "\nID: "
            + name
            + ", Leg: "
            + legname
            + "\nMismatch between SC num. of XLS SC Column ("
            + str(user_scnum)
            + ") & \nSCs with values in Swing/"
            + "Stance columns ("
            + str(total_scnum)
            + ")!"
            + "\nWe used all valid swing/stance entries ("
            + str(total_scnum)
            + ")."
            + "\nCheck your table."
        )
        print(this_message)
        write_issues_to_textfile(this_message, info)

    # ...........................  idxs to all_cycles  .................................
    # use value we found, loop over all runs, loop over SCs within a run
    # => all_cycles is a list of len = number of runs with lists within that having
    #    len = number of SCs of each run
    all_cycles = [[] for r in range(len(run_scnums))]  # run, not user!
    for r in range(len(run_scnums)):
        all_cycles[r] = [[None, None] for s in range(run_scnums[r])]
    for r, run in enumerate(runs):
        run_row = runs[runs == run].index
        for s in range(run_scnums[r]):
            if s == 0:
                start_col = SCdf.columns.get_loc(SWINGSTART_COL)
                end_col = SCdf.columns.get_loc(STANCEEND_COL)
            else:
                # str(s) because colnames match s for s>0 (& @else our loop starts @ 1)!
                start_col = SCdf.columns.get_loc(SWINGSTART_COL + "." + str(s))
                end_col = SCdf.columns.get_loc(STANCEEND_COL + "." + str(s))
            # time as floats
            start_in_s = float(SCdf.iloc[run_row, start_col].values[0])
            end_in_s = float(SCdf.iloc[run_row, end_col].values[0])
            # see if we are rounding to fix inaccurate user input
            # => account for python's float precision leading to inaccuracies
            # => two important steps here (sanity_check_vals only used for these checks)
            # 1. round to 10th decimal to fix python making
            #    3211.999999999999999995 out of 3212
            sanity_check_start = round(start_in_s * sampling_rate, 10)
            sanity_check_end = round(end_in_s * sampling_rate, 10)
            # 2. comparing abs(sanity check vals) to 1e-7 just to be 1000% sure
            if (abs(sanity_check_start % 1) > 1e-7) | (
                abs(sanity_check_end % 1) > 1e-7
            ):
                round_message = (
                    "\n***********\n! WARNING !\n***********\n"
                    + "SC latencies of "
                    + str(start_in_s)
                    + "s to "
                    + str(end_in_s)
                    + "s were not provided in units of the frame rate!"
                    + "\nWe thus use the previous possible frame(s)."
                    + "\nDouble check if this worked as expected or fix annotation table!"
                )
                print(round_message)
                write_issues_to_textfile(round_message, info)
            # assign to all_cycles (note int() rounds down!)
            all_cycles[r][s][0] = int(start_in_s * sampling_rate)
            all_cycles[r][s][1] = int(end_in_s * sampling_rate)
            # check if we are in data bounds
            if (all_cycles[r][s][0] in data.index) & (
                all_cycles[r][s][1] in data.index
            ):
                pass
            else:
                all_cycles[r][s] = [None, None]  # so they can be cleaned later
                this_message = (
                    "\n***********\n! WARNING !\n***********\n"
                    + legname
                    + " leg"
                    + " - Run #"
                    + str(r + 1)
                    + " - SC #"
                    + str(s + 1)
                    + " is out of data-bounds - Skipping!"
                )
                print(this_message)
                write_issues_to_textfile(this_message, info)
    # ............................  clean all_cycles  ..................................
    # check if we skipped latencies because they were out of data-bounds
    all_cycles = check_cycle_out_of_bounds(all_cycles)
    if all_cycles:  # can be None if all SCs were out of bounds
        # check if there are any duplicates (e.g., SC2's start-lat == SC1's end-lat)
        all_cycles = check_cycle_duplicates(all_cycles)
        # check if user input progressively later latencies
        all_cycles = check_cycle_order(all_cycles, info, legname)
        # check that none of the joints used for angles have the same value within SCs
        all_cycles = check_different_angle_joint_coords(all_cycles, data, info, cfg)
        # NOTE for future self
        # => If you are considering to remove empty lists from run_cycles note that we
        #    need them for the case of Runs 1 & 3 having SCs but 2 not having any
        # => Otherwise SC-level plots arent plotted correctly (ie Run3 could easily
        #    look like Run2 - we need an empty subplot panel for run2 and thus an empty
        #    list!)
    return all_cycles


def check_different_angle_joint_coords(all_cycles, data, info, cfg):
    """Check if none of the joints used for angle computations later have equal values (since this would lead to math.domain errors due to floating point precision)"""

    # Note 1
    # ------
    # In theory, I could fix this programatically in the add_angle function, but I feel
    # like joint-coords should not often be exactly equal like this in a meaningful way
    # We can still change it in the future.

    # Note 2
    # ------
    # It is important that I loop over legs, because I have to ensure that angle-joints
    # are different for each leg separately - even though the SC is just w.r.t. one leg,
    # both legs' angles are computed later (and would lead to an error if equal)

    # unpack
    angles = cfg["angles"]

    clean_cycles = None
    for r, run_cycles in enumerate(all_cycles):
        for c, cycle in enumerate(run_cycles):
            cycle = check_a_single_cycle_for_joint_coords(
                cycle, angles, data, r, c, info
            )
            if cycle:  # if cycle was not valid (equal-joint-coords) this returns None
                if clean_cycles is None:  # if first valid cycle: list of empty lists
                    clean_cycles = [[] for s in range(len(all_cycles))]
                clean_cycles[r].append(cycle)
    return clean_cycles


def check_a_single_cycle_for_joint_coords(cycle, angles, data, r, c, info):
    """This checks a single cycle: across both legs and all angle configs. Removing the cycle if we had equal joint-coords anywhere!"""
    for legname in LEGS:  # loop over legs
        for a in range(len(angles["name"])):  # loop over different angle configs
            # first, prepare a dict that has only the data of this leg/angle combo
            this_angle_data = {"name": [], "lower_joint": [], "upper_joint": []}
            for key in this_angle_data.keys():
                this_joint = angles[key][a]
                if this_joint + "Y" in data.columns:
                    this_angle_data[key] = np.array(
                        [data[this_joint + "Y"], data[this_joint + "Z"]]
                    )
                else:  # we know that this joint is in data (because of a check in prep)
                    this_angle_data[key] = np.array(
                        [
                            data[this_joint + ", " + legname + " Y"],
                            data[this_joint + ", " + legname + " Z"],
                        ]
                    )
            # now check if any of the joints have the same coord at any idx
            for idx in range(cycle[0], cycle[1] + 1):  # within this SC only
                if (
                    np.array_equal(
                        this_angle_data["name"][:, idx],
                        this_angle_data["lower_joint"][:, idx],
                    )
                    or np.array_equal(
                        this_angle_data["name"][:, idx],
                        this_angle_data["upper_joint"][:, idx],
                    )
                    or np.array_equal(
                        this_angle_data["lower_joint"][:, idx],
                        this_angle_data["upper_joint"][:, idx],
                    )
                ):
                    this_message = (
                        "\n***********\n! WARNING !\n***********\n"
                        + f"Run #{r + 1} - SC #{c + 1} has equal {legname.upper()} "
                        + "joint coordinates at "
                        + f"{round(data[TIME_COL][idx],4)}s:"
                        + "\n\nAngle - [Y Z]:\n"
                        + angles["name"][a]
                        + " - "
                        + str(this_angle_data["name"][:, idx])
                        + "\nLower joint: "
                        + angles["lower_joint"][a]
                        + " - "
                        + str(this_angle_data["lower_joint"][:, idx])
                        + "\nUpper joint: "
                        + angles["upper_joint"][a]
                        + " - "
                        + str(this_angle_data["upper_joint"][:, idx])
                        + "\nRemoving the SC from "
                        + f"{round(data[TIME_COL][cycle[0]], 4)}-"
                        + f"{round(data[TIME_COL][cycle[1]], 4)}s"
                    )
                    print(this_message)
                    write_issues_to_textfile(this_message, info)
                    return None  # removes this SC
    return cycle  # if we never returned None, this SC is valid


# ..............................  helper functions  ....................................
def check_cycle_out_of_bounds(all_cycles):
    """Check if user provided SC latencies that were not in video/data bounds"""
    clean_cycles = None
    for r, run_cycles in enumerate(all_cycles):
        for c, cycle in enumerate(run_cycles):
            # below checks if values are any type of int (just in case int-type should
            # for some super random reason change...)
            if isinstance(cycle[0], (int, np.integer)) & isinstance(
                cycle[1], (int, np.integer)
            ):
                if clean_cycles is None:
                    clean_cycles = [[] for s in range(len(all_cycles))]
                clean_cycles[r].append(cycle)
    return clean_cycles


def check_cycle_duplicates(all_cycles):
    """Check if there are any duplicate SC latencies.
    This would break our plotting functions, which use .loc on all_steps_data - thus,
    all indices of all_cycles have to be unique. If any duplicates found, add one
    datapoint to the start latency.
    """
    for r, run_cycles in enumerate(all_cycles):
        for c, cycle in enumerate(run_cycles):
            if c > 0:
                if cycle[0] == run_cycles[c - 1][1]:
                    all_cycles[r][c][0] += 1
    return all_cycles


def check_cycle_order(all_cycles, info, legname):
    """Check if user input flawed SC latencies

    Two cases
    1. Start latency earlier than end latency of previous SC
    2. End latency earlier then start latency of current SC
    """

    clean_cycles = [[] for s in range(len(all_cycles))]
    current_max_time = 0  # outside of for loops so it persists across runs
    for r, run_cycles in enumerate(all_cycles):
        for c, cycle in enumerate(run_cycles):
            if cycle[0] > current_max_time:
                if cycle[1] > cycle[0]:
                    clean_cycles[r].append(cycle)  # only append if both tests passed
                    current_max_time = cycle[1]
                else:
                    this_message = (
                        "\n***********\n! WARNING !\n***********\n"
                        + legname
                        + " - Run #"
                        + str(r + 1)
                        + " - SC #"
                        + str(c + 1)
                        + " has a later start than end latency - Skipping!"
                    )
                    print(this_message)
                    write_issues_to_textfile(this_message, info)
            else:
                this_message = (
                    "\n***********\n! WARNING !\n***********\n"
                    + legname
                    + " - Run #"
                    + str(r + 1)
                    + " - SC #"
                    + str(c + 1)
                    + " has an earlier start than previous SC's end latency - Skipping!"
                )
                print(this_message)
                write_issues_to_textfile(this_message, info)
    return clean_cycles


def check_stepcycles(all_cycles, info):
    """Check results of SC extraction. Cancel everything if None found!"""
    name = info["name"]
    # case 1 - valid SCs for both legs
    if isinstance(all_cycles["left"], list) and isinstance(all_cycles["right"], list):
        return all_cycles
    # case 2 - no valid SCs for left leg
    elif (all_cycles["left"] is None) and isinstance(all_cycles["right"], list):
        this_message = (
            "\n***********\n! ERROR !\n***********\n"
            + "\nID: "
            + name
            + "\nNo valid SCs found for LEFT leg!"
        )
        write_issues_to_textfile(this_message, info)
        print(this_message)
        return all_cycles
    # case 3 - no valid SCs for right leg
    elif isinstance(all_cycles["left"], list) and (all_cycles["right"] is None):
        this_message = (
            "\n***********\n! ERROR !\n***********\n"
            + "\nID: "
            + name
            + "\nNo valid SCs found for RIGHT leg!"
        )
        write_issues_to_textfile(this_message, info)
        print(this_message)
        return all_cycles
    # case 4 - no valid SCs for either leg
    elif (all_cycles["left"] is None) & (all_cycles["right"] is None):
        this_message = (
            "\n******************\n! CRITICAL ERROR !\n******************\n"
            + "\nID: "
            + name
            + "\nSkipped because no valid SCs found for any leg!"
        )
        print(this_message)
        write_issues_to_textfile(this_message, info)
        return  # in this case, abort everything (returns None to main)



import os
import json
import tkinter as tk


# ............................ gui config file operations  ............................
def update_config_file(
    results,
    cfg,
    autogaita_folder_path,
    config_file_name,
    list_vars,
    dict_vars,
    tk_str_vars,
    tk_bool_vars,
):
    """updates the gui_config.json file with this runs parameters"""
    # transform tkVars into normal strings and bools
    output_dicts = [{}, {}]
    for i in range(len(output_dicts)):
        if i == 0:
            # in case update_config_file is called before results is defined
            # as in the creation of the exit_button in the dlc_gui() function
            # the results dict of the last run is used and only cfg is updated
            if results == "results dict not defined yet":
                # runwindow = None as we dont need the tk.Vars to refer to a specific window
                input_dict = extract_results_from_json_file(
                    None,
                    autogaita_folder_path,
                    config_file_name,
                    tk_str_vars,
                    tk_bool_vars,
                )
            else:
                input_dict = results
        elif i == 1:
            input_dict = cfg
        for key in input_dict.keys():
            if key in list_vars:
                # if list of strings, initialise output empty list and get & append vals
                output_dicts[i][key] = []
                for list_idx in range(len(input_dict[key])):
                    output_dicts[i][key].append(input_dict[key][list_idx].get())
            elif key in dict_vars:
                # if dict of list of strings, initialise as empty dict and assign stuff
                # key = "angles" or other DICT_VAR
                # inner_key = "name" & "lower_ / upper_joint"
                # list_idx = idx of list of strings of inner_key
                output_dicts[i][key] = {}
                for inner_key in input_dict[key]:
                    output_dicts[i][key][inner_key] = []
                    for list_idx in range(len(input_dict[key][inner_key])):
                        output_dicts[i][key][inner_key].append(
                            input_dict[key][inner_key][list_idx].get()
                        )
            else:
                output_dicts[i][key] = input_dict[key].get()

    # merge the two configuration dictionaries
    configs_list = [output_dicts[0], output_dicts[1]]  # 0 = results, 1 = cfg, see above
    # write the configuration file
    with open(
        os.path.join(autogaita_folder_path, config_file_name), "w"
    ) as config_json_file:
        json.dump(configs_list, config_json_file, indent=4)


def extract_cfg_from_json_file(
    window,
    autogaita_folder_path,
    config_file_name,
    list_vars,
    dict_vars,
    tk_str_vars,
    tk_bool_vars,
):
    """loads the cfg dictionary from the config file"""
    # load the configuration file
    with open(
        os.path.join(autogaita_folder_path, config_file_name), "r"
    ) as config_json_file:
        # config_json contains list with 0 -> result and 1 -> cfg data
        last_runs_cfg = json.load(config_json_file)[1]

    cfg = {}
    # assign values to the cfg dict
    for key in last_runs_cfg.keys():
        if key in tk_bool_vars:
            cfg[key] = tk.BooleanVar(window, last_runs_cfg[key])
        elif key in list_vars:
            cfg[key] = []
            for entry in last_runs_cfg[key]:
                cfg[key].append(tk.StringVar(window, entry))
        elif key in dict_vars:
            cfg[key] = {}
            for subkey in last_runs_cfg[key]:
                cfg[key][subkey] = []
                for entry in last_runs_cfg[key][subkey]:
                    cfg[key][subkey].append(tk.StringVar(window, entry))
        elif key in tk_str_vars:  # Integers are also saved as strings
            cfg[key] = tk.StringVar(window, last_runs_cfg[key])
    return cfg


def extract_results_from_json_file(
    window, autogaita_folder_path, config_file_name, tk_str_vars, tk_bool_vars
):
    """loads the results dictionary from the config file"""

    # load the configuration file
    with open(
        os.path.join(autogaita_folder_path, config_file_name), "r"
    ) as config_json_file:
        # config_json contains list with 0 -> result and 1 -> cfg data
        last_runs_results = json.load(config_json_file)[0]

    results = {}
    for key in last_runs_results.keys():
        if key in tk_str_vars:
            results[key] = tk.StringVar(window, last_runs_results[key])
        elif key in tk_bool_vars:
            results[key] = tk.BooleanVar(window, last_runs_results[key])

    return results



# %% imports
from autogaita.resources.utils import (
    write_issues_to_textfile,
    standardise_primary_joint_coordinates,
)
import os
import shutil
import json
import pandas as pd
import numpy as np

# %% constants
from autogaita.resources.constants import (
    ISSUES_TXT_FILENAME,
    CONFIG_JSON_FILENAME,
    TIME_COL,
)
from autogaita.universal3D.universal3D_constants import (
    LEGS_COLFORMAT,
)

# %% workflow step #1 - preparation


# ................................  main function  .....................................
def some_prep(info, folderinfo, cfg):
    """Preparation of the data for later analyses"""
    # ............................  unpack stuff  ......................................
    name = info["name"]
    results_dir = info["results_dir"]
    postname_string = folderinfo["postname_string"]
    sampling_rate = cfg["sampling_rate"]
    flip_gait_direction = cfg["flip_gait_direction"]
    standardise_z_at_SC_level = cfg["standardise_z_at_SC_level"]
    analyse_average_y = cfg["analyse_average_y"]
    standardise_y_coordinates = cfg["standardise_y_coordinates"]
    standardise_z_to_a_joint = cfg["standardise_z_to_a_joint"]
    coordinate_standardisation_xls = cfg["coordinate_standardisation_xls"]
    tracking_software = "Universal 3D"  # hardcoded for this toolbox

    # .............................  move data  ........................................
    # => see if we can delete a previous runs results folder if existant. if not, it's a
    #    bit ugly since we only update results if filenames match...
    # => for example if angle acceleration not wanted in current run, but was stored in
    #    previous run, the previous run's figure is in the folder
    # => inform the user and leave this as is
    if os.path.exists(results_dir):
        try:
            shutil.rmtree(results_dir)
            move_data_to_folders(info, folderinfo)
        except OSError:
            move_data_to_folders(info, folderinfo)
            unable_to_rm_resdir_error = (
                "\n***********\n! WARNING !\n***********\n"
                + "Unable to remove previous Results subfolder of ID: "
                + name
                + "!\n Results will only be updated if filenames match!"
            )
            print(unable_to_rm_resdir_error)
            write_issues_to_textfile(unable_to_rm_resdir_error, info)
    else:
        move_data_to_folders(info, folderinfo)

    # .......  initialise Issues.txt & quick check for file existence  .................
    issue_txt_path = os.path.join(results_dir, ISSUES_TXT_FILENAME)
    if os.path.exists(issue_txt_path):
        os.remove(issue_txt_path)
    # read data
    # => note that this catches xls and xlsx files. can do this since we previously
    #    handle cases of both being @ root_dir in move_data_to_folders function
    # => using empty data df & empty string for error handling
    data = pd.DataFrame(data=None)
    data_duplicate_error = ""
    for filename in os.listdir(results_dir):
        if name + postname_string + ".xls" in filename:
            if data.empty:
                try:
                    data = pd.read_excel(os.path.join(results_dir, filename))
                except:
                    data = pd.read_excel(
                        os.path.join(results_dir, filename), engine="openpyxl"
                    )
            else:
                data_duplicate_error = (
                    "\n******************\n! CRITICAL ERROR !\n******************\n"
                    + "Two DATA xls-files found for "
                    + name
                    + "!\nPlease ensure your root directory only has one "
                    + "datafile per ID"
                )

    # ............................  import data  .......................................
    import_error_message = ""  # prep stuff for error handling
    if data_duplicate_error:
        import_error_message += data_duplicate_error
    if data.empty:
        import_error_message += (
            "\n******************\n! CRITICAL ERROR !\n******************\n"
            + "Unable to load an xls/xlsx file with data for "
            + name
            + "!\nTry again!"
        )
    if import_error_message:  # see if there was any issues with import, if so: stop
        print(import_error_message)
        write_issues_to_textfile(import_error_message, info)
        return (None, None)

    # ................  final data checks, conversions & additions  ....................
    # IMPORTANT
    # ---------
    # MAIN TESTS OF USER-INPUT VALIDITY OCCUR HERE!
    cfg = test_and_expand_cfg(data, cfg, info)
    if cfg == (None, None):  # joints were empty
        return (None, None)  # => return the tuple bc. some prep returns 2 variables
    joints = cfg["joints"]
    angles = cfg["angles"]
    # important to unpack to vars and not to cfg since cfg is overwritten in multiruns!
    y_standardisation_joint = cfg["y_standardisation_joint"][0]
    z_standardisation_joint = cfg["z_standardisation_joint"][0]
    # store config json file @ group path
    # !!! NU - do this @ ID path
    group_path = results_dir.split(name)[0]
    config_json_path = os.path.join(group_path, CONFIG_JSON_FILENAME)
    config_vars_to_json = {
        "sampling_rate": sampling_rate,
        "flip_gait_direction": flip_gait_direction,
        "standardise_z_at_SC_level": standardise_z_at_SC_level,
        "analyse_average_y": analyse_average_y,
        "standardise_y_coordinates": standardise_y_coordinates,
        "y_standardisation_joint": y_standardisation_joint,
        "standardise_z_to_a_joint": standardise_z_to_a_joint,
        "z_standardisation_joint": z_standardisation_joint,
        "coordinate_standardisation_xls": coordinate_standardisation_xls,
        "joints": joints,
        "angles": angles,
        "tracking_software": tracking_software,
    }
    # note - using "w" will overwrite/truncate file, thus no need to remove it if exists
    with open(config_json_path, "w") as config_json_file:
        json.dump(config_vars_to_json, config_json_file, indent=4)

    # Check if data has some col saying "Time" in any form of capitalisation and if so
    # make sure it's capitalised
    if any(col.lower() == TIME_COL.lower() for col in data.columns):
        data.columns = [
            col.capitalize() if col.lower() == TIME_COL.lower() else col
            for col in data.columns
        ]

    # Annoying thing 1 of our simi data: there were two Time = 0s, we take the second
    # (i.e. last)
    if TIME_COL in data.columns and len(np.where(data[TIME_COL] == 0)[0]) > 1:
        real_start_idx = np.where(data[TIME_COL] == 0)[0][-1]
        data = data.iloc[real_start_idx:, :]
        data.index = range(len(data))  # update index

    # Important: either create time col if not present or if present set its values
    data[TIME_COL] = data.index * (1 / sampling_rate)
    cols = [TIME_COL] + [col for col in data.columns if col != TIME_COL]  # 1st col
    data = data[cols]

    # Annoying thing 2 of our simi data: it sometimes does weird things with their data
    # (e.g., storing 99cm as 0,99 and 1 metre 10cm something something as 101.222.333
    # or so) - catch it
    try:
        data = data.astype(float)
    except:
        unable_to_convert_message = (
            "\n******************\n! CRITICAL ERROR !\n******************\n"
            + "Unable to convert data to numbers for "
            + name
            + ".\nThis is likely due to inconsistent or wrong decimal separators in "
            + "some columns."
            + "\nTry again!"
        )
        print(unable_to_convert_message)
        write_issues_to_textfile(unable_to_convert_message, info)
        raise ValueError(unable_to_convert_message)

    # Standardise y columns to be positive & afterwards save global y_max for flipping
    y_cols = [col for col in data.columns if col.endswith("Y")]
    global_Y_min = min(data[y_cols].min())
    if global_Y_min < 0:
        data[y_cols] += abs(global_Y_min)
    global_Y_max = max(data[y_cols].max())

    # Standardise all Z columns to global Z minimum or a user-provided joint
    z_cols = [col for col in data.columns if col.endswith("Z")]  # Find all Z cols
    if standardise_z_to_a_joint:
        # Note in the test-function we ensured that this was provided in the correct
        # "Foot, left" format and handled trailing spaces being too much or missing
        z_min = data[z_standardisation_joint + "Z"].min()
    else:
        z_min = min(data[z_cols].min())  # Compute global z min
    data[z_cols] -= z_min  # Subtract either joint's or global z min from all Z cols

    # Standardise all user-chosen "joint" coordinates dividing them by a fixed decimal
    # => IMPORTANT: This must be done AFTER all y-/z-standardisations above!
    # => note this function has proper errors raised if things are wrongly configured
    #    (so we don't have to return (None, None)...)
    if len(coordinate_standardisation_xls) > 0:
        data, cfg = standardise_primary_joint_coordinates(
            data, tracking_software, info, cfg
        )
    return data, global_Y_max


# ..............................  helper functions  ....................................


def move_data_to_folders(info, folderinfo):
    """Copy data to new results_dir"""
    # unpack
    name = info["name"]
    results_dir = info["results_dir"]
    root_dir = folderinfo["root_dir"]
    postname_string = folderinfo["postname_string"]
    # create res dir (has to be first thing because of using issues.txt file here)
    os.makedirs(results_dir)
    # check if there is an xls and an xlsx table for name
    xls_flag = False
    xlsx_flag = False
    for filename in os.listdir(root_dir):
        if not postname_string:
            if filename.endswith(name + ".xls"):
                xls_flag = True
            elif filename.endswith(name + ".xlsx"):
                xlsx_flag = True
        else:
            if filename.endswith(name + postname_string + ".xls"):
                xls_flag = True
            elif filename.endswith(name + postname_string + ".xlsx"):
                xlsx_flag = True
    if (xls_flag is True) & (xlsx_flag is True):
        two_table_warning = (
            "\n***********\n! WARNING !\n***********\n"
            + "We found an xls as well as an xlsx file for ID: "
            + name
            + "!\nWe will use the .xlsx file -  fix & re-run if this is unwanted!"
        )
        print(two_table_warning)
        write_issues_to_textfile(two_table_warning, info)
    # move correct xls(x) file to it
    for filename in os.listdir(root_dir):
        if name + postname_string + ".xls" in filename:
            if filename.endswith(".xls"):
                if xlsx_flag is False:
                    shutil.copy2(
                        os.path.join(root_dir, filename),
                        os.path.join(results_dir, filename),
                    )
            else:
                shutil.copy2(
                    os.path.join(root_dir, filename),
                    os.path.join(results_dir, filename),
                )


def test_and_expand_cfg(data, cfg, info):
    """Test some important cfg variables and add new ones based on them

    Procedure
    ---------
    Check that no strings are empty
    Check that all strings end with a space character
    Check that all features are present in the dataset
    Add plot_joints & direction_joint
    Make sure to set dont_show_plots to True if Python is not in interactive mode
    """

    # run the 3 tests first
    for cfg_key in ["angles", "joints"]:
        cfg[cfg_key] = check_and_fix_cfg_strings(data, cfg, cfg_key, info)

    # add plot_joints - used for average plots & stick diagram
    joints = cfg["joints"]
    if cfg["plot_joint_number"] > len(joints):  # 1) joints to plot in detail
        this_message = (
            "\n***********\n! WARNING !\n***********\n"
            + "You asked us to plot more joints than available!"
            + "\nNumber of joints to plot: "
            + str(cfg["plot_joint_number"])
            + "\nNumber of selected joints: "
            + str(len(joints))
            + "\n\nWe'll just plot the most we can :)"
        )
        write_issues_to_textfile(this_message, info)
        print(this_message)
        cfg["plot_joints"] = joints
    else:
        cfg["plot_joints"] = joints[: cfg["plot_joint_number"]]

    # add direction_joint - used to determine gait direction
    if not joints:  # no valid string after above cleaning
        no_joint_message = (
            "\n******************\n! CRITICAL ERROR !\n******************\n"
            + "After testing your joint names, no valid joint was left to "
            + "perform gait direction checks on.\nPlease make sure that at least one "
            + "joint is provided & try again!"
        )
        write_issues_to_textfile(no_joint_message, info)
        print(no_joint_message)
        return (None, None)  # returning tuple bc. some_prep returns 2 variables
    # this way of setting direction joint works because we prev. checked its validity
    if joints[0] + "Y" in data.columns:
        cfg["direction_joint"] = joints[0] + "Y"
    else:
        cfg["direction_joint"] = joints[0] + LEGS_COLFORMAT[0] + "Y"

    # test y/z standardisation joints if needed
    broken_standardisation_joint = ""
    if cfg["standardise_y_coordinates"]:
        cfg["y_standardisation_joint"] = check_and_fix_cfg_strings(
            data, cfg, "y_standardisation_joint", info
        )
        if not cfg["y_standardisation_joint"]:
            broken_standardisation_joint += "y"
    if cfg["standardise_z_to_a_joint"]:
        cfg["z_standardisation_joint"] = check_and_fix_cfg_strings(
            data, cfg, "z_standardisation_joint", info
        )
        if not cfg["z_standardisation_joint"]:
            if broken_standardisation_joint:
                broken_standardisation_joint += " & z"
            else:
                broken_standardisation_joint += "z"
    if broken_standardisation_joint:
        no_standardisation_joint_message = (
            "\n******************\n! CRITICAL ERROR !\n******************\n"
            + "After testing your standardisation joints we found an issue with "
            + broken_standardisation_joint
            + "-coordinate standardisation joint."
            + "\n Cancelling AutoGaitA - please try again!"
        )
        write_issues_to_textfile(no_standardisation_joint_message, info)
        print(no_standardisation_joint_message)
        return (None, None)  # returning tuple bc. some_prep returns 2 variables

    return cfg


def check_and_fix_cfg_strings(data, cfg, cfg_key, info):
    """Check and fix strings in our joint & angle lists so that:
    1) They don't include empty strings
    2) All strings end with the space character (since we do string + "Z")
    3) All strings are valid columns of the coordinate dataset

    Note - THIS IS NOT AND SHOULD NOT BE IDENTICAL TO 2D VERSIONS OF THIS FUNCTION!
    ----
    This is a bit more involved here than in _dlc since we have to check whether
    a given string is body-specific (e.g., Midfoot, left Z) or not (e.g., Pelvis Z)
    => Based on this the string of the feature itself has to either end or not end with
       a space character (see doc of check_data_column_names for more info)
    """

    # work on this variable (we return to cfg[key] outside of here)
    string_variable = cfg[cfg_key]

    # easy checks: lists
    if type(string_variable) is list:
        string_variable = [string for string in string_variable if string]
        # note that below local func. DOES NOT CLEAN the list - it only corrects space
        # characters and returns invalid_idxs that we'll use next to clean the list
        string_variable, invalid_joint_idxs = check_data_column_names(
            data, string_variable, cfg_key=cfg_key
        )
        # if user gave us joints we didnt find, write, print & save error message
        if invalid_joint_idxs:  # if none found, func returns an empty list (falsey!)
            # backup dirty variable
            dirty_string_variable = string_variable
            # now clean the list
            string_variable = [
                item
                for i, item in enumerate(string_variable)
                if i not in invalid_joint_idxs
            ]
            # now print & save erroneous & remaining joints
            joint_error = (
                "\n***********\n! WARNING !\n***********\n"
                + "Unable to find all "
                + cfg_key
                + " joints/key points you entered!"
                + "\n Note that body-side specific y/z standardisation joints must be "
                + " provided as such, e.g. 'Foot, left'"
                + "\n\nInvalid and thus removed were:"
            )
            for idx in invalid_joint_idxs:
                joint_error += "\n" + dirty_string_variable[idx]
            joint_error += "\n\nWe removed these and will analyse the remaining ones:"
            for string in string_variable:
                joint_error += "\n" + string
            joint_error += (
                "\n\nNote that capitalisation matters."
                + "\nIf you are running a batch analysis, we'll use this updated cfg "
                + "throughout.\nCheck out the config.json file for the full cfg used."
            )
            print(joint_error)
            write_issues_to_textfile(joint_error, info)

    # things are more involved for angle dicts
    elif type(string_variable) is dict:
        # 1) test if the lists of all keys are equally long, if not throw out last idxs
        key_lengths = [len(string_variable[key]) for key in string_variable.keys()]
        if not all(key_length == key_lengths[0] for key_length in key_lengths):
            min_length = min(key_lengths)
            for key in string_variable:  # remove invalid idxs
                string_variable[key] = string_variable[key][:min_length]
            key_length_mismatch_message = (  # inform user
                "\n***********\n! WARNING !\n***********\n"
                + "\nLength-mismatch in angle configuration!"
                + "\nCheck angles' name/upper/lower-joint entries."
                + "\nOnly processing first "
                + str(min_length)
                + " entries."
            )
            print(key_length_mismatch_message)
            write_issues_to_textfile(key_length_mismatch_message, info)
        # 2) check if all strings in the angle dict are valid columns
        invalid_angletrio_message = ""
        invalid_idxs = []  # these idxs hold across the 3 keys of our angles-dict
        for key in string_variable:
            this_keys_missing_strings = ""
            string_variable[key], this_keys_invalid_idxs = check_data_column_names(
                data, string_variable[key]
            )
            if this_keys_invalid_idxs:  # if none found this is an empty list (falsey!)
                invalid_idxs.append(this_keys_invalid_idxs)
                if not this_keys_missing_strings:  # first occurance
                    this_keys_missing_strings += "\nAngle's " + key + " key: "
                for idx in this_keys_invalid_idxs:
                    this_keys_missing_strings += (
                        string_variable[key][idx] + " (#" + str(idx + 1) + ") / "
                    )
                # string concat outside of idx-forloop above please
                invalid_angletrio_message += this_keys_missing_strings
        # if we have to remove idxs from all keys of our angles dict
        if invalid_angletrio_message:
            # print and save
            angles_warning = (
                "\n***********\n! WARNING !\n***********\n"
                + "Unable to find all "
                + cfg_key
                + " joints/key points you entered!"
                + "\n\nInvalid and thus removed were:"
                + invalid_angletrio_message
            )
            print(angles_warning)
            write_issues_to_textfile(angles_warning, info)
            # clean the dict
            if type(invalid_idxs[0]) is list:  # if needed, flatten to a list of values
                invalid_idxs = [item for sublist in invalid_idxs for item in sublist]
            invalid_idxs = list(set(invalid_idxs))  # remove duplicates
            for key in string_variable:
                string_variable[key] = [
                    string
                    for i, string in enumerate(string_variable[key])
                    if i not in invalid_idxs
                ]
            # print & save info about the remaining ("clean") angles dict
            clean_angles_message = "\nRemaining angle configuration is:"
            for i in range(len(string_variable["name"])):
                clean_angles_message += "\n\nAngle #" + str(i + 1)
                clean_angles_message += "\nName: " + string_variable["name"][i]
                clean_angles_message += (
                    "\nLower Joint: " + string_variable["lower_joint"][i]
                )
                clean_angles_message += (
                    "\nUpper Joint: " + string_variable["upper_joint"][i]
                )
            clean_angles_message += (
                "\n\nNote that capitalisation matters."
                + "\nIf you are running a batch analysis, we'll use this updated cfg "
                + "throughout\nCheck out the config.json file for the full cfg used."
            )
            print(clean_angles_message)
            write_issues_to_textfile(clean_angles_message, info)

    return string_variable


def check_data_column_names(data, string_variable, **kwargs):
    """Checks if user-input is present in data columns and changes space-characters
    accordingly
    => Bodyside-specific colnames have to end WITHOUT a space (because we concat
       "name" + ", leg " + "Z" - so leg ends with a space)
    => Bodyside-nonspecific colnames have to end WITH a space (because we concat "name "
       + "Z" so name has to end with a space)

    IMPORTANT NOTE
    --------------
    This function does NOT REMOVE invalid strings from the list!!! It is by design that
    you have to remove those outside of this function using invalid_joint_idx.
    HOWEVER IT DOES fix the space-character thingy for those strings THAT ARE IN THE
    DATASET!
    """
    cfg_key = kwargs.get("cfg_key", None)
    invalid_joint_idxs = []
    for s, string in enumerate(string_variable):
        # if a string is ending with a space, see if it is bodyside-specific
        # if that's the case, remove the space
        if string.endswith(" "):
            if string + "Y" in data.columns:
                pass
            # this "elif" condition applies to joints/direction joints, it's different
            # for y/z-standardisation, which is checked further down
            elif string[:-1] + LEGS_COLFORMAT[0] + "Y" in data.columns:
                string_variable[s] = string[:-1]
            else:
                invalid_joint_idxs.append(s)
        # if a string is not ending with a space, see if it should
        else:
            # this "if" is similar to "elif" above
            if string + LEGS_COLFORMAT[0] + "Y" in data.columns:
                pass
            elif string + " Y" in data.columns:
                string_variable[s] = string + " "
            else:
                invalid_joint_idxs.append(s)
        # SPECIAL CASE for standardisation joints which must be provided side-specific
        # if corresponding joint is.
        # => ONLY DOING THE FOLLOWING FOR SIDE-SPECIFIC JOINTS!
        # => if string ends with a space ("Foot, left ") it should lead to valid joint
        #    if "Y" added. flag invalid joint if it doesn't
        # => if string doesn't end with a space and would be valid if it did, add space
        if cfg_key in ("y_standardisation_joint", "z_standardisation_joint"):
            # if string DOES NOT END (!) with a space "Foot, left" make sure to test it
            # with a space because LEGS_COLFORMAT ends with spaces
            if not string.endswith(" "):  # NOTE THE "NOT" HERE!
                test_string = string + " "
                if any(leg in test_string for leg in LEGS_COLFORMAT):
                    if test_string + "Y" not in data.columns:
                        invalid_joint_idxs.append(s)
                    else:
                        # if it's valid, add the space to string_variable
                        string_variable[s] = string + " "
            else:
                # the "else" of this "if" means that string_variable was given
                # correctly and can stay as is
                if any(leg in string for leg in LEGS_COLFORMAT):
                    if string + "Y" not in data.columns:
                        invalid_joint_idxs.append(s)
            # this happens if user gave "Foot" for example without the
            # side-specific identifier (e.g. "Foot, left") - catch it
            if string + "Y" not in data.columns and string + " Y" not in data.columns:
                if s not in invalid_joint_idxs:  # no duplicates!
                    invalid_joint_idxs.append(s)

    return string_variable, invalid_joint_idxs



from autogaita.resources.utils import (
    standardise_primary_joint_coordinates,
    compute_angle,
    write_angle_warning,
)
from autogaita.common2D.common2D_1_preparation import some_prep as some_prep_2D
from autogaita.universal3D.universal3D_1_preparation import some_prep as some_prep_3D
from autogaita.common2D.common2D_2_sc_extraction import extract_stepcycles
from autogaita.common2D.common2D_3_analysis import analyse_and_export_stepcycles
import os
import math
import pandas as pd
import pandas.testing as pdt
import pytest


# %%...........................  2D GaitA fixtures  ....................................
@pytest.fixture
def extract_2D_info(tmp_path):
    info = {}
    info["mouse_num"] = 15
    info["run_num"] = 3
    info["name"] = "ID " + str(info["mouse_num"]) + " - Run " + str(info["run_num"])
    info["results_dir"] = os.path.join(tmp_path, info["name"])
    return info


@pytest.fixture
def extract_2D_folderinfo():
    folderinfo = {}
    folderinfo["root_dir"] = "tests/test_data/dlc_data"
    folderinfo["sctable_filename"] = (
        "correct_annotation_table.xlsx"  # has to be an excel file
    )
    folderinfo["data_string"] = "SIMINewOct"
    folderinfo["beam_string"] = "BeamTraining"
    folderinfo["premouse_string"] = "Mouse"
    folderinfo["postmouse_string"] = "25mm"
    folderinfo["prerun_string"] = "run"
    folderinfo["postrun_string"] = "6DLC"
    return folderinfo


@pytest.fixture
def extract_2D_cfg():
    cfg = {}
    cfg["sampling_rate"] = 100
    cfg["subtract_beam"] = False
    cfg["dont_show_plots"] = True
    cfg["convert_to_mm"] = False  # false!
    cfg["pixel_to_mm_ratio"] = 3.76
    cfg["x_sc_broken_threshold"] = 200
    cfg["y_sc_broken_threshold"] = 50
    cfg["x_acceleration"] = True
    cfg["angular_acceleration"] = True
    cfg["save_to_xls"] = True
    cfg["bin_num"] = 25
    cfg["plot_SE"] = True
    cfg["standardise_y_at_SC_level"] = False
    cfg["standardise_y_to_a_joint"] = True
    cfg["y_standardisation_joint"] = ["Knee"]
    cfg["plot_joint_number"] = 3
    cfg["color_palette"] = "viridis"
    cfg["legend_outside"] = True
    cfg["invert_y_axis"] = True
    cfg["flip_gait_direction"] = False
    cfg["analyse_average_x"] = True
    cfg["standardise_x_coordinates"] = True
    cfg["x_standardisation_joint"] = ["Hind paw tao"]
    cfg["coordinate_standardisation_xls"] = ""
    cfg["hind_joints"] = ["Hind paw tao", "Ankle", "Knee", "Hip", "Iliac Crest"]
    cfg["fore_joints"] = [
        "Front paw tao ",
        "Wrist ",
        "Elbow ",
        "Lower Shoulder ",
        "Upper Shoulder ",
    ]
    cfg["beam_col_left"] = ["BeamLeft"]  # BEAM_COL_LEFT & _RIGHT must be lists of len=1
    cfg["beam_col_right"] = ["BeamRight"]
    cfg["beam_hind_jointadd"] = ["Tail base ", "Tail center ", "Tail tip "]
    cfg["beam_fore_jointadd"] = ["Nose ", "Ear base "]
    cfg["angles"] = {
        "name": ["Ankle ", "Knee ", "Hip "],
        "lower_joint": ["Hind paw tao ", "Ankle ", "Knee "],
        "upper_joint": ["Knee ", "Hip ", "Iliac Crest "],
    }
    return cfg


# %%...........................  3D GaitA fixtures  ....................................
@pytest.fixture
def extract_3D_info(tmp_path):
    info = {}
    info["name"] = "TestSubject"
    info["results_dir"] = os.path.join(tmp_path, info["name"])
    return info


@pytest.fixture
def extract_3D_folderinfo():
    folderinfo = {}
    folderinfo["root_dir"] = "tests/test_data/universal3D_data/test_data/"
    folderinfo["sctable_filename"] = "SC Latency Table"
    folderinfo["postname_string"] = ""
    return folderinfo


@pytest.fixture
def extract_3D_cfg():
    cfg = {}
    cfg["sampling_rate"] = 100
    cfg["dont_show_plots"] = True
    cfg["y_acceleration"] = True
    cfg["angular_acceleration"] = True
    cfg["bin_num"] = 25
    cfg["plot_SE"] = True
    cfg["standardise_z_at_SC_level"] = True
    cfg["standardise_z_to_a_joint"] = False
    cfg["z_standardisation_joint"] = ["Midfoot, left"]
    cfg["plot_joint_number"] = 7
    cfg["legend_outside"] = True
    cfg["flip_gait_direction"] = False
    cfg["color_palette"] = "viridis"
    cfg["analyse_average_y"] = False
    cfg["standardise_y_coordinates"] = True
    cfg["y_standardisation_joint"] = ["Midfoot, left"]
    cfg["coordinate_standardisation_xls"] = ""
    cfg["joints"] = ["Midfoot", "Ankle", "Knee", "Hip", "Pelvis "]
    cfg["angles"] = {
        "name": ["Ankle", "Knee", "Hip"],
        "lower_joint": ["Midfoot", "Ankle", "Knee"],
        "upper_joint": ["Knee", "Hip", "Pelvis "],
    }
    return cfg


# %% .................................  tests  .........................................


def test_compute_angle():
    # Test case 1: Basic case
    joint1 = [0, 0]
    joint2 = [1, 0]
    joint3 = [0, 1]
    expected_angle = 90
    result, _ = compute_angle(joint1, joint2, joint3)
    assert math.isclose(result, expected_angle)
    # Test case 2: two joints are equal - this won't happen in autogaita because of the
    # test in extract_stepcycles but I want to make sure that the function returns
    # broken=True correctly
    joint1 = [5, 5]
    joint2 = [2, 2]
    joint3 = [2, 2]
    expected_angle = None
    result, broken = compute_angle(joint1, joint2, joint3)
    assert result == 0
    assert broken is True


def test_write_angle_warning(extract_2D_cfg, extract_2D_info):
    # prep: remove existing Issues.txt
    results_dir = extract_2D_info["results_dir"]
    if not os.path.exists(results_dir):
        os.makedirs(results_dir)
    issues_path = os.path.join(results_dir, "Issues.txt")
    if os.path.exists(issues_path):
        os.remove(issues_path)
    # prep: more vars
    step = pd.DataFrame()
    step["Time"] = [0.1, 0.2, 0.3]
    step["dummy_coord"] = [1, 2, 3]  # need this too otherwise stuff breaks
    angles = extract_2D_cfg["angles"]
    a = 0
    broken_angle_idxs = [0, 2]
    # run
    write_angle_warning(step, a, angles, broken_angle_idxs, extract_2D_info)
    # assert
    with open(issues_path, "r") as f:
        issues = f.read()
    assert (
        "Angle: Ankle" in issues  # bc. of extract_2D_cfg
        and "Lower Joint: Hind paw tao" in issues
        and "Upper Joint: Knee" in issues
        and "Cycle-time: 0.1-0.3s" in issues
    )
    # run again - test legname works as expected
    os.remove(issues_path)  # first remove previous textfile
    write_angle_warning(
        step, a, angles, broken_angle_idxs, extract_2D_info, legname="left"
    )
    with open(issues_path, "r") as f:
        issues = f.read()
    assert "Leg: left" in issues


def test_correct_coordinate_standardisation(
    extract_2D_info,
    extract_2D_folderinfo,
    extract_2D_cfg,
    extract_3D_info,
    extract_3D_folderinfo,
    extract_3D_cfg,
):
    for tracking_software in ["DLC", "Universal 3D"]:
        # prep vars
        if tracking_software == "DLC":
            info = extract_2D_info
            folderinfo = extract_2D_folderinfo
            cfg = extract_2D_cfg
        else:
            info = extract_3D_info
            folderinfo = extract_3D_folderinfo
            cfg = extract_3D_cfg
        # run respective some_prep functions to get dfs
        if tracking_software == "DLC":
            # unstandardised data
            cfg["coordinate_standardisation_xls"] = ""
            unstandardised_data = some_prep_2D(tracking_software, info, folderinfo, cfg)
            # standardised data
            cfg["coordinate_standardisation_xls"] = (
                "tests/test_data/utils/Correct DLC CoordStand Table.xlsx"
            )
            standardised_data = some_prep_2D(tracking_software, info, folderinfo, cfg)
        elif tracking_software == "Universal 3D":
            # unstandardised data
            cfg["coordinate_standardisation_xls"] = ""
            unstandardised_data = some_prep_3D(info, folderinfo, cfg)[0]  # tuple!
            # standardised data
            cfg["coordinate_standardisation_xls"] = (
                "tests/test_data/utils/Correct Universal 3D CoordStand Table.xlsx"
            )
            standardised_data, global_Y_max = some_prep_3D(info, folderinfo, cfg)
        # revert standardisation
        reverted_data = standardised_data.copy()
        standardisation_df = pd.read_excel(
            cfg["coordinate_standardisation_xls"]
        ).astype(str)
        if tracking_software == "DLC":
            condition = (standardisation_df["ID"] == str(info["mouse_num"])) & (
                standardisation_df["Run"] == str(info["run_num"])
            )
        elif tracking_software == "Universal 3D":
            condition = standardisation_df["ID"] == info["name"]
        standardisation_value = float(
            standardisation_df.loc[condition, "Standardisation Value"]
        )
        if tracking_software == "DLC":
            cols_to_revert = [
                col
                for col in reverted_data.columns
                if (not col.endswith("likelihood"))
                and any([joint in col for joint in cfg["hind_joints"]])
            ]
        elif tracking_software == "Universal 3D":
            cols_to_revert = [
                col
                for col in reverted_data.columns
                if any([joint in col for joint in cfg["joints"]])
            ]
        reverted_data[cols_to_revert] *= standardisation_value

        # compare dataframes
        pd.testing.assert_frame_equal(
            reverted_data, unstandardised_data, check_exact=False, check_dtype=False
        )


def test_angles_are_unaffected_by_coordinate_standardisation(
    extract_2D_info, extract_2D_folderinfo, extract_2D_cfg
):
    # prep: run dlc_main's first 3 steps to get dfs with angles
    # 1) for unstandardised data
    data = some_prep_2D("DLC", extract_2D_info, extract_2D_folderinfo, extract_2D_cfg)
    all_cycles = extract_stepcycles(
        "DLC", data, extract_2D_info, extract_2D_folderinfo, extract_2D_cfg
    )
    unstandardised_results = analyse_and_export_stepcycles(
        data, all_cycles, extract_2D_info, extract_2D_cfg
    )
    # 2) for standardised data
    extract_2D_cfg["coordinate_standardisation_xls"] = (
        "autogaita/resources/Coordinate Standardisation Table Template.xlsx"
    )
    data = some_prep_2D("DLC", extract_2D_info, extract_2D_folderinfo, extract_2D_cfg)
    all_cycles = extract_stepcycles(
        "DLC", data, extract_2D_info, extract_2D_folderinfo, extract_2D_cfg
    )
    standardised_results = analyse_and_export_stepcycles(
        data, all_cycles, extract_2D_info, extract_2D_cfg
    )
    # compare angles
    cols_to_compare = [
        col
        for col in unstandardised_results["average_data"].columns
        if col.endswith("Angle")
    ]
    pdt.assert_frame_equal(
        unstandardised_results["average_data"][cols_to_compare],
        standardised_results["average_data"][cols_to_compare],
    )


# Parameterized test for error cases
@pytest.mark.parametrize(
    "xls_path, expected_error",
    [
        (
            "autogaita/resources/This CoordStand Table is Missing.xlsx",
            "No coordinate standardisation xls file found at:",
        ),
        (
            "tests/test_data/utils/This CoordStand Table has wrong columns.xlsx",
            "does not have the correct column names",
        ),
        (
            "tests/test_data/utils/This CoordStand Table has wrong Run.xlsx",
            "Unable to find",
        ),
        (
            "tests/test_data/utils/This CoordStand Table has multiple Names.xlsx",
            "Found multiple entries for",
        ),
        (
            "tests/test_data/utils/This CoordStand Table doesn't have a float as standval.xlsx",
            "Unable to convert standardisation value for",
        ),
        (
            "tests/test_data/utils/This CoordStand Table has a value smaller than 1.xlsx",
            "smaller than 1!",
        ),
    ],
)
def test_standardisation_xls_error_cases(
    extract_2D_info, extract_2D_folderinfo, extract_2D_cfg, xls_path, expected_error
):
    # prep: remove existing Issues.txt
    results_dir = extract_2D_info["results_dir"]
    issues_path = os.path.join(results_dir, "Issues.txt")
    if os.path.exists(issues_path):
        os.remove(issues_path)

    # set the xls path in the config & run some_prep which will run the standardisation
    extract_2D_cfg["coordinate_standardisation_xls"] = xls_path
    with pytest.raises(Exception):
        data = some_prep_2D(
            "DLC", extract_2D_info, extract_2D_folderinfo, extract_2D_cfg
        )

    # assert the error message - inform about what error failed if it did
    with open(issues_path, "r") as f:
        issues = f.read()
    assert (
        expected_error in issues
    ), f"Expected error '{expected_error}' not found in Issues.txt"


def test_angle_joints_not_in_joints(
    extract_2D_info, extract_2D_folderinfo, extract_2D_cfg
):
    # prep cfg stuff
    cfg = extract_2D_cfg.copy()
    cfg["coordinate_standardisation_xls"] = (
        "tests/test_data/utils/Correct DLC CoordStand Table.xlsx"
    )
    cfg["hind_joints"] = ["Ankle", "Knee", "Hip"]
    cfg["angles"] = {  # hind paw is to be removed (not in cfg["joints"]!)
        "name": ["Hind paw tao", "Knee", "Knee", "Knee"],
        "lower_joint": ["Ankle", "Hind paw tao", "Ankle", "Ankle"],
        "upper_joint": ["Hip", "Hip", "Hind paw tao", "Hip"],
    }
    # prep: remove existing Issues.txt
    results_dir = extract_2D_info["results_dir"]
    issues_path = os.path.join(results_dir, "Issues.txt")
    if os.path.exists(issues_path):
        os.remove(issues_path)
    # run (not just some prep, because we want the cfg, too)
    data = some_prep_2D("DLC", extract_2D_info, extract_2D_folderinfo, cfg)
    data, cfg = standardise_primary_joint_coordinates(data, "DLC", extract_2D_info, cfg)
    # assert error
    with open(issues_path, "r") as f:
        issues = f.read()
    assert "We will update your cfg (!)" in issues
    # assert updated cfg
    assert cfg["angles"] == {
        "name": ["Knee "],
        "lower_joint": ["Ankle "],
        "upper_joint": ["Hip "],
    }



# %% imports
from autogaita.resources.utils import (
    write_issues_to_textfile,
    standardise_primary_joint_coordinates,
)
from autogaita.common2D.common2D_constants import FILE_ID_STRING_ADDITIONS
import os
import shutil
import json
import pandas as pd
import numpy as np
import h5py

# %% constants
from autogaita.resources.constants import (
    ISSUES_TXT_FILENAME,
    TIME_COL,
    CONFIG_JSON_FILENAME,
)
from autogaita.common2D.common2D_constants import (
    DIRECTION_DLC_THRESHOLD,
)

# %% workflow step #1 - preparation


def some_prep(tracking_software, info, folderinfo, cfg):
    """Preparation of the data & cfg file for dlc analyses"""

    # .........................  unpack & prep stuff  ..................................
    # IMPORTANT
    # ---------
    # file_type_string handles all places (like move_data_to_folders) where it makes a
    # difference if we are processing DLC or SLEAP files
    if tracking_software == "DLC":
        file_type_string = ".csv"
    elif tracking_software == "SLEAP":
        file_type_string = ".h5"
    # DON'T unpack (joint) cfg-keys that are tested later by check_and_expand_cfg
    name = info["name"]
    results_dir = info["results_dir"]
    data_string = folderinfo["data_string"]
    beam_string = folderinfo["beam_string"]
    sampling_rate = cfg["sampling_rate"]
    subtract_beam = cfg["subtract_beam"]
    convert_to_mm = cfg["convert_to_mm"]
    pixel_to_mm_ratio = cfg["pixel_to_mm_ratio"]
    x_sc_broken_threshold = cfg["x_sc_broken_threshold"]
    y_sc_broken_threshold = cfg["y_sc_broken_threshold"]
    standardise_y_at_SC_level = cfg["standardise_y_at_SC_level"]
    invert_y_axis = cfg["invert_y_axis"]
    flip_gait_direction = cfg["flip_gait_direction"]
    analyse_average_x = cfg["analyse_average_x"]
    standardise_x_coordinates = cfg["standardise_x_coordinates"]
    standardise_y_to_a_joint = cfg["standardise_y_to_a_joint"]
    coordinate_standardisation_xls = cfg["coordinate_standardisation_xls"]

    # .............................  move data  ........................................
    # => slightly different for DLC or SLEAP (see local functions below)
    # => see if we can delete a previous runs results folder if existant. if not, it's a
    #    bit ugly since we only update results if filenames match...
    # => for example if angle acceleration not wanted in current run, but was stored in
    #    previous run, the previous run's figure is in the folder
    # => inform the user and leave this as is
    if os.path.exists(results_dir):
        try:
            shutil.rmtree(results_dir)
            move_data_to_folders(tracking_software, file_type_string, info, folderinfo)
        except OSError:
            move_data_to_folders(tracking_software, file_type_string, info, folderinfo)
            unable_to_rm_resdir_error = (
                "\n***********\n! WARNING !\n***********\n"
                + "Unable to remove previous Results subfolder of ID: "
                + name
                + "!\n Results will only be updated if filenames match!"
            )
            print(unable_to_rm_resdir_error)
            write_issues_to_textfile(unable_to_rm_resdir_error, info)
    else:
        move_data_to_folders(tracking_software, file_type_string, info, folderinfo)

    # .......  initialise Issues.txt & quick check for file existence  .................
    # Issues.txt - delete if saved in a previous run
    issues_txt_path = os.path.join(results_dir, ISSUES_TXT_FILENAME)
    if os.path.exists(issues_txt_path):
        os.remove(issues_txt_path)
    # read data & beam
    if not os.listdir(results_dir):
        no_files_error = (
            "\n******************\n! CRITICAL ERROR !\n******************\n"
            + "Unable to identify ANY RELEVANT FILES for "
            + name
            + "!\nThis is likely due to issues with unique file name identifiers.. "
            + "check capitalisation!"
        )
        write_issues_to_textfile(no_files_error, info)
        print(no_files_error)
        return

    # ............................  import data  .......................................
    datadf = pd.DataFrame(data=None)  # prep stuff for error handling
    datadf_duplicate_error = ""
    if subtract_beam:
        if data_string == beam_string:
            beam_and_data_string_error_message = (
                "\n******************\n! CRITICAL ERROR !\n******************\n"
                + "Your data & baseline (beam) identifiers ([G] in our "
                + "file  naming convention) are identical. "
                + "\nNote that they must be different! \nTry again"
            )
            write_issues_to_textfile(beam_and_data_string_error_message, info)
            return
        beamdf = pd.DataFrame(data=None)
        beamdf_duplicate_error = ""
    for filename in os.listdir(results_dir):  # import
        if filename.endswith(file_type_string):
            if data_string in filename:
                if datadf.empty:
                    if tracking_software == "DLC":
                        datadf = pd.read_csv(os.path.join(results_dir, filename))
                    elif tracking_software == "SLEAP":
                        datadf = h5_to_df(results_dir, filename)
                else:
                    datadf_duplicate_error = (
                        "\n******************\n! CRITICAL ERROR !\n******************\n"
                        + f"Two DATA {file_type_string}-files found for {name}"
                        + "!\nPlease ensure your root directory only has one datafile "
                        + "per video!"
                    )
            if subtract_beam:
                if beam_string in filename:
                    if beamdf.empty:
                        if tracking_software == "DLC":
                            beamdf = pd.read_csv(os.path.join(results_dir, filename))
                        elif tracking_software == "SLEAP":
                            beamdf = h5_to_df(results_dir, filename)
                    else:
                        beamdf_duplicate_error = (
                            "\n******************\n! CRITICAL ERROR !\n***************"
                            + f"***\nTwo BEAM {file_type_string}-files found for {name}"
                            + "!\nPlease ensure your root directory only has one "
                            + "beamfile per video!"
                        )
    # handle import errors
    # => append to empty strings to handle multiple issues at once seemlessly
    import_error_message = ""
    if datadf_duplicate_error:
        import_error_message += datadf_duplicate_error
    if datadf.empty:
        import_error_message += (  # if pd didn't raise errors but dfs still empty
            "\n******************\n! CRITICAL ERROR !\n******************\n"
            + f"Unable to load a DATA {file_type_string} file for {name}"
            + "!\nTry again!"
        )
    if subtract_beam:
        if beamdf_duplicate_error:
            import_error_message += beamdf_duplicate_error
        if beamdf.empty:
            import_error_message += (
                "\n******************\n! CRITICAL ERROR !\n******************\n"
                + f"Unable to load a BEAM {file_type_string} file for {name}"
                + "!\nTry again!"
            )
    if import_error_message:  # see if there was any issues with import, if so: stop
        print(import_error_message)
        write_issues_to_textfile(import_error_message, info)
        return

    # ....  finalise import: rename cols, get rid of unnecessary elements, floatit  ....
    if tracking_software == "DLC":  # need to prep DLC dfs
        datadf = prepare_DLC_df(datadf)
        if subtract_beam:  # beam df
            beamdf = prepare_DLC_df(beamdf)
    if subtract_beam:
        data = pd.concat([datadf, beamdf], axis=1)
    else:
        data = datadf.copy(deep=True)

    # ................  final data checks, conversions & additions  ....................
    # IMPORTANT - MAIN TESTS OF USER-INPUT VALIDITY OCCUR HERE!
    # => UNPACK VARS FROM CFG THAT ARE TESTED BY check_and_expand HERE, NOT EARLIER!
    cfg = check_and_expand_cfg(data, cfg, info)
    if cfg is None:  # some critical error occured
        return
    hind_joints = cfg["hind_joints"]
    fore_joints = cfg["fore_joints"]
    angles = cfg["angles"]
    beam_hind_jointadd = cfg["beam_hind_jointadd"]
    beam_fore_jointadd = cfg["beam_fore_jointadd"]
    direction_joint = cfg["direction_joint"]
    # important to unpack to vars and not to cfg since cfg is overwritten in multiruns!
    x_standardisation_joint = cfg["x_standardisation_joint"][0]
    y_standardisation_joint = cfg["y_standardisation_joint"][0]
    # store config json file @ group path
    # !!! NU - do this @ mouse path!
    group_path = results_dir.split(name)[0]
    config_json_path = os.path.join(group_path, CONFIG_JSON_FILENAME)
    config_vars_to_json = {
        "sampling_rate": sampling_rate,
        "convert_to_mm": convert_to_mm,
        "pixel_to_mm_ratio": pixel_to_mm_ratio,
        "x_sc_broken_threshold": x_sc_broken_threshold,
        "y_sc_broken_threshold": y_sc_broken_threshold,
        "standardise_y_at_SC_level": standardise_y_at_SC_level,
        "flip_gait_direction": flip_gait_direction,
        "analyse_average_x": analyse_average_x,
        "standardise_x_coordinates": standardise_x_coordinates,
        "x_standardisation_joint": x_standardisation_joint,
        "standardise_y_to_a_joint": standardise_y_to_a_joint,
        "y_standardisation_joint": y_standardisation_joint,
        "coordinate_standardisation_xls": coordinate_standardisation_xls,
        "hind_joints": hind_joints,
        "fore_joints": fore_joints,
        "angles": angles,
        "tracking_software": tracking_software,
    }
    # note - using "w" will overwrite/truncate file, thus no need to remove it if exists
    with open(config_json_path, "w") as config_json_file:
        json.dump(config_vars_to_json, config_json_file, indent=4)
    # a little test to see if columns make sense, i.e., same number of x/y/likelihood
    x_col_count = len([c for c in data.columns if c.endswith(" x")])
    y_col_count = len([c for c in data.columns if c.endswith(" y")])
    likelihood_col_count = "N/A because SLEAP"  # initialise so message doesnt break
    if tracking_software == "DLC":
        likelihood_col_count = len(
            [c for c in data.columns if c.endswith(" likelihood")]
        )
        col_count_condition = x_col_count == y_col_count == likelihood_col_count
    elif tracking_software == "SLEAP":
        col_count_condition = x_col_count == y_col_count
    if col_count_condition:
        pass
    else:
        cols_are_weird_message = (
            "\n***********\n! WARNING !\n***********\n"
            + "We detected an unequal number of columns ending with x, y or "
            + "likelihood!\nCounts were:\n"
            + "x: "
            + str(x_col_count)
            + ", y: "
            + str(y_col_count)
            + ", likelihood: "
            + str(likelihood_col_count)
            + "!\n\n"
            + "We continue with the analysis but we strongly suggest you have another "
            + "look at your dataset, this should not happen.\n"
        )
        print(cols_are_weird_message)
        write_issues_to_textfile(cols_are_weird_message, info)
    # if wanted: fix that deeplabcut & SLEAP inverse y
    if invert_y_axis:
        for col in data.columns:
            if col.endswith(" y"):
                data[col] = data[col] * -1
    # if we don't have a beam to subtract, standardise y to a joint's or global ymin = 0
    if not subtract_beam:
        y_min = float("inf")
        y_cols = [col for col in data.columns if col.endswith("y")]
        if standardise_y_to_a_joint:
            y_min = data[y_standardisation_joint + "y"].min()
        else:
            y_min = data[y_cols].min().min()
        data[y_cols] -= y_min
    # quick warning if cfg is set to not flip gait direction but to standardise x
    if not flip_gait_direction and standardise_x_coordinates:
        message = (
            "\n***********\n! WARNING !\n***********\n"
            + "You are standardising x-coordinates without standardising the direction "
            + "of gait (e.g. all walking from right to left)."
            + "\nThis can be correct if you are doing things like treadmill walking "
            + "but can lead to unexpected behaviour otherwise!"
            + "\nMake sure you know what you are doing!"
        )
        print(message)
        write_issues_to_textfile(message, info)
    # check gait direction & DLC file validity
    data = check_gait_direction(
        tracking_software, data, direction_joint, flip_gait_direction, info
    )
    if data is None:  # this means DLC file is broken
        return
    # subtract the beam from the joints to standardise y
    # => bc. we simulate that all mice run from left to right, we can write:
    #     (note that we also flip beam x columns, but never y-columns!)
    # => & bc. we multiply y values by *-1 earlier, it's a neg_num - - neg_num
    #    pushing it towards zero.
    # => using list(set()) to ensure that we don't have duplicate values (if users
    #    should have provided them in both cfg vars by misstake)
    # => beam_col_left and right is provided by users
    if subtract_beam:
        # note beam_col_left/right are always lists in cfg!
        beam_col_left = cfg["beam_col_left"][0]
        beam_col_right = cfg["beam_col_right"][0]
        for joint in list(set(hind_joints + beam_hind_jointadd)):
            data[joint + "y"] = data[joint + "y"] - data[beam_col_left + "y"]
        for joint in list(set(fore_joints + beam_fore_jointadd)):
            data[joint + "y"] = data[joint + "y"] - data[beam_col_right + "y"]
        data.drop(columns=list(beamdf.columns), inplace=True)  # beam not needed anymore
    # add Time
    data[TIME_COL] = data.index * (1 / sampling_rate)
    # reorder the columns we added
    cols = [TIME_COL, "Flipped"]
    data = data[cols + [c for c in data.columns if c not in cols]]
    # ------------------------------------------------------------------------
    #                             IMPORTANT
    # Next two things must be the last that are done in this function, since
    # joint-standardisation must be done after beam-subtraction & non-beam-height-stand
    # and pixel-standardisation must be done after joint-standardisation!
    # ------------------------------------------------------------------------
    # 1) standardise all primary joint (!) coordinates
    # => all dimensions are divided by a fixed user-provided value
    # => note this function has proper errors raised if things should be wrongly
    #    configured by user!
    if len(coordinate_standardisation_xls) > 0:
        data, cfg = standardise_primary_joint_coordinates(
            data, tracking_software, info, cfg
        )
    # 2) convert pixels to millimeters
    if convert_to_mm:
        for column in data.columns:
            if column.endswith((" x", " y")):  # if tuple: endswith any of these
                data[column] = data[column] / pixel_to_mm_ratio
    return data


# ........................  tracking software specific helpers  ........................
def h5_to_df(results_dir, filename):
    """Convert a SLEAP h5 file to the pandas dataframe used in gaita"""
    df = pd.DataFrame(data=None)
    with h5py.File(os.path.join(results_dir, filename), "r") as f:
        locations = f["tracks"][:].T
        node_names = [n.decode() for n in f["node_names"][:]]
        df.index = np.arange(np.shape(locations)[0])
        for node_idx, node_name in enumerate(node_names):
            for c, coord in enumerate(["x", "y"]):
                df[node_name + " " + coord] = locations[:, node_idx, c, 0]
    return df


def prepare_DLC_df(df, separator=" "):
    """Prepare the DLC dataframe after loading w.r.t. column names & df-index
    Note
    ----
    separator is used in universal3D_datafile_preparation
    """
    new_column_strings = list()  # data df
    for j in range(df.shape[1]):
        new_column_strings.append(df.iloc[0, j] + separator + df.iloc[1, j])
    df.columns = new_column_strings
    # next lines indices are because: scorer row becomes the column, bodypart row is row
    # 0, coords row is row 1 and we thus include row 2 onwards. col 1 onwards is obvious
    df = df.iloc[2:, 1:]
    df.index = range(len(df))
    df = df.astype(float)
    return df


# ...............................  generic helpers  ....................................


def move_data_to_folders(tracking_software, file_type_string, info, folderinfo):
    """Find files, copy data, video, beamdata & beamvideo to new results_dir"""
    # unpack
    results_dir = info["results_dir"]
    postmouse_string = folderinfo["postmouse_string"]
    postrun_string = folderinfo["postrun_string"]
    os.makedirs(results_dir)  # important to do this outside of loop!
    # check if user forgot some underscores or dashes in their filenames
    # => two levels of string additions for two post FILE-ID strings
    # => in theory if the user has some strange cases in which this double forloop
    # would be true twice (because one file is called -6DLC and another is called _6DLC
    # for some reason) it will break after the first time and always ignore the second
    # one - keep this in mind if it should come up but it should be very unlikely
    for mouse_string_addition in FILE_ID_STRING_ADDITIONS:
        candidate_postmouse_string = mouse_string_addition + postmouse_string
        for run_string_addition in FILE_ID_STRING_ADDITIONS:
            candidate_postrun_string = run_string_addition + postrun_string
            found_it = check_this_filename_configuration(
                tracking_software,
                file_type_string,
                info,
                folderinfo,
                candidate_postmouse_string,
                candidate_postrun_string,
                results_dir,
            )
            if found_it:  # if our search was successful, stop searching and continue
                break


def check_this_filename_configuration(
    tracking_software,
    file_type_string,
    info,
    folderinfo,
    postmouse_string,
    postrun_string,
    results_dir,
):
    # unpack
    name = info["name"]
    mouse_num = info["mouse_num"]
    run_num = info["run_num"]
    root_dir = folderinfo["root_dir"]
    data_string = folderinfo["data_string"]
    beam_string = folderinfo["beam_string"]
    premouse_string = folderinfo["premouse_string"]
    prerun_string = folderinfo["prerun_string"]
    whichvideo = ""  # initialise
    found_it = False
    # handle leading zeros that we identified previously - if none convert nums to str
    if "leading_mouse_num_zeros" in info.keys():
        mouse_num = info["leading_mouse_num_zeros"] + str(mouse_num)
    else:
        mouse_num = str(mouse_num)
    if "leading_run_num_zeros" in info.keys():
        run_num = info["leading_run_num_zeros"] + str(run_num)
    else:
        run_num = str(run_num)
    for filename in os.listdir(root_dir):
        # the following condition is True for data & beam csv
        if (
            (premouse_string + mouse_num + postmouse_string in filename)
            and (prerun_string + run_num + postrun_string in filename)
            and (filename.endswith(file_type_string))
        ):
            found_it = True
            # Copy the csv/h5 file to the new subfolder
            shutil.copy2(
                os.path.join(root_dir, filename), os.path.join(results_dir, filename)
            )
            # Check if there is a video and if so copy it too (only for DLC atm)
            if tracking_software == "DLC":
                vidname = filename[:-4] + "_labeled.mp4"
                vidpath = os.path.join(root_dir, vidname)
                if os.path.exists(vidpath):
                    shutil.copy2(vidpath, os.path.join(results_dir, vidname))
                else:
                    if data_string in vidname:
                        whichvideo = "Data"
                    elif beam_string in vidname:
                        whichvideo = "Beam"
                    this_message = (
                        "\n***********\n! WARNING !\n***********\n"
                        + "No "
                        + whichvideo
                        + "video for "
                        + name
                        + "!"
                    )
                    print(this_message)
                    write_issues_to_textfile(this_message, info)
    return found_it


def check_gait_direction(
    tracking_software, data, direction_joint, flip_gait_direction, info
):
    """Check direction of gait - reverse it if needed

    Note for DLC
    ------------
    Also using this check to check for DLC files being broken
    flip_gait_direction is only used after the check for DLC files being broken
    """

    # AN IMPORTANT NOTE
    # -----------------
    # The general approach is different for DLC and much more involved. Besides the
    # check for DLC files being broken, the np.median() approach used for SLEAP (in the
    # end) of this function cannot be used too well for DLC IMO because DLC has those
    # huge numbers when estimating without being confident that would mess with the
    # medians

    # INITIALISE FLIPPED COLUMN
    data["Flipped"] = False

    # DLC APPROACH
    if tracking_software == "DLC":
        enterframe = 0
        idx = 0
        flip_error_message = ""
        while enterframe == 0:  # first find out when mouse was in the video frame.
            if (
                np.mean(data[direction_joint + "likelihood"][idx : idx + 5])
                > DIRECTION_DLC_THRESHOLD
            ):  # +5 to increase conf.
                enterframe = idx + 5
            idx += 1
            if (idx > len(data)) | (enterframe > len(data)):
                flip_error_message += (
                    "\n******************\n! CRITICAL ERROR !\n******************\n"
                    + "Unable to determine gait direction!"
                    + "\nThis hints a critical issue with DLC tracking, e.g., likelihood "
                    + "\ncolumns being low everywhere or tables being suspiciously short!"
                    + "\nTo be sure, we cancel everything here."
                    + "\nPlease check your input DLC csv files for correctness & try again!"
                )
                break
        leaveframe = 0
        idx = 1
        while leaveframe == 0:  # see where mouse left frame (same logic from back)
            if (
                np.mean(data[direction_joint + "likelihood"][-idx - 5 : -idx])
                > DIRECTION_DLC_THRESHOLD
            ):
                leaveframe = len(data) - idx - 5
            idx += 1
            if idx > len(data):
                if not flip_error_message:
                    flip_error_message += (
                        "\n******************\n! CRITICAL ERROR !\n******************\n"
                        + "Unable to determine gait direction!"
                        + "\nThis hints a critical issue with DLC tracking, e.g., "
                        + "likelihood \ncolumns being low everywhere or tables being "
                        + "suspiciously short!\nTo be sure, we cancel everything here."
                        + "\nPlease check your input DLC csv files for correctness & try "
                        + "again!"
                    )
                break
        if flip_error_message:
            write_issues_to_textfile(flip_error_message, info)
            print(flip_error_message)
            return
        if (
            data[direction_joint + "x"][enterframe]
            > data[direction_joint + "x"][leaveframe]
        ):  # i.e.: right to left
            # simulate that mouse ran from left to right (only if user wants it)
            if flip_gait_direction:
                data = flip_mouse_body(data, info)
                data["Flipped"] = True

    # SLEAP APPROACH
    elif tracking_software == "SLEAP":
        # beloow if condition means that the mouse ran from right to left
        # => in this case we flip
        if np.median(data[direction_joint + "x"][: len(data) // 2]) > np.median(
            data[direction_joint + "x"][len(data) // 2 :]
        ):
            if flip_gait_direction:
                data = flip_mouse_body(data, info)
                data["Flipped"] = True

    # RETURN DATA
    return data


def check_and_expand_cfg(data, cfg, info):
    """Test some important cfg variables and add new ones based on them

    Procedure
    ---------
    Check that no strings are empty
    Check that all strings end with a space character
    Check that all features are present in the dataset
    Add plot_joints & direction_joint
    Make sure to set dont_show_plots to True if Python is not in interactive mode
    If users subtract a beam, set normalise @ sc level to False
    String-checks for standardisation joints
    """

    # run the tests first
    # => note that beamcols & standardisation joints are tested separately further down.
    for cfg_key in [
        "angles",
        "hind_joints",
        "fore_joints",
    ]:
        cfg[cfg_key] = check_and_fix_cfg_strings(data, cfg, cfg_key, info)

    # add plot_joints - used for average plots & stick diagram
    hind_joints = cfg["hind_joints"]
    if cfg["plot_joint_number"] > len(hind_joints):
        fix_plot_joints_message = (
            "\n***********\n! WARNING !\n***********\n"
            + "You asked us to plot more hind limb joints than available!"
            + "\nNumber of joints to plot: "
            + str(cfg["plot_joint_number"])
            + "\nNumber of selected hindlimb joints: "
            + str(len(hind_joints))
            + "\n\nWe'll just plot the most we can :)"
        )
        write_issues_to_textfile(fix_plot_joints_message, info)
        print(fix_plot_joints_message)
        cfg["plot_joints"] = hind_joints
    else:
        cfg["plot_joints"] = hind_joints[: cfg["plot_joint_number"]]

    # check hindlimb joints & add direction_joint - used to determine gait direction
    if not hind_joints:  # no valid string after above cleaning
        no_hind_joint_message = (
            "\n******************\n! CRITICAL ERROR !\n******************\n"
            + "After testing your hind limb joint names, no valid joint was left to "
            + "perform gait direction checks on.\nPlease make sure that at least one "
            + "hind limb joint is provided & try again!"
        )
        write_issues_to_textfile(no_hind_joint_message, info)
        print(no_hind_joint_message)
        return
    cfg["direction_joint"] = hind_joints[0]

    # if subtracting beam, check identifier-strings & that beam colnames were valid.
    if cfg["subtract_beam"]:
        # first, let's check the strings
        # => note beamcols are lists of len=1 bc. of check function
        for cfg_key in [
            "beam_col_left",
            "beam_col_right",
            "beam_hind_jointadd",
            "beam_fore_jointadd",
        ]:
            cfg[cfg_key] = check_and_fix_cfg_strings(data, cfg, cfg_key, info)
        beam_col_error_message = (
            "\n******************\n! CRITICAL ERROR !\n******************\n"
            + "It seems like you want to standardise heights to a baseline (beam)."
            + "\nUnfortunately we were unable to find the y-columns you listed in "
            + "your beam-file.\nPlease try again.\nInvalid beam side(s) was/were:"
        )
        beam_error = False  # check 3 possible cases
        if (cfg["beam_col_left"]) and (not cfg["beam_col_right"]):
            beam_error = True
            beam_col_error_message += "\n right beam!"
        elif (not cfg["beam_col_left"]) and (cfg["beam_col_right"]):
            beam_error = True
            beam_col_error_message += "\n left beam!"
        elif (not cfg["beam_col_left"]) and (not cfg["beam_col_right"]):
            beam_error = True
            beam_col_error_message += "\n both beams!"
        if beam_error:  # if any case was True, stop everything
            write_issues_to_textfile(beam_col_error_message, info)
            print(beam_col_error_message)
            return
    # never standardise @ SC level if user subtracted a beam
    if cfg["subtract_beam"] is True:
        cfg["standardise_y_at_SC_level"] = False
    # test x/y standardisation joints if needed
    broken_standardisation_joint = ""
    if cfg["standardise_x_coordinates"]:
        cfg["x_standardisation_joint"] = check_and_fix_cfg_strings(
            data, cfg, "x_standardisation_joint", info
        )
        if not cfg["x_standardisation_joint"]:
            broken_standardisation_joint += "x"
    if cfg["standardise_y_to_a_joint"]:
        cfg["y_standardisation_joint"] = check_and_fix_cfg_strings(
            data, cfg, "y_standardisation_joint", info
        )
        if not cfg["y_standardisation_joint"]:
            if broken_standardisation_joint:
                broken_standardisation_joint += " & y"
            else:
                broken_standardisation_joint += "y"
    if broken_standardisation_joint:
        no_standardisation_joint_message = (
            "\n******************\n! CRITICAL ERROR !\n******************\n"
            + "After testing your standardisation joints we found an issue with "
            + broken_standardisation_joint
            + "-coordinate standardisation joint."
            + "\n Cancelling AutoGaitA - please try again!"
        )
        write_issues_to_textfile(no_standardisation_joint_message, info)
        print(no_standardisation_joint_message)
        return

    return cfg


def check_and_fix_cfg_strings(data, cfg, cfg_key, info):
    """Check and fix strings in our joint & angle lists so that:
    1) They don't include empty strings
    2) All strings end with the space character
       => Important note: strings should never have the coordinate in them (since we do
          string + "y" for example throughout this code)
    3) All strings are valid columns of the DLC dataset
       => Note that x_standardisation_joint is tested against ending with "x" - rest
          against "y"
    """

    # work on this variable (we return to cfg[key] outside of here)
    string_variable = cfg[cfg_key]

    # easy checks: lists
    if type(string_variable) is list:
        string_variable = [s for s in string_variable if s]
        string_variable = [s if s.endswith(" ") else s + " " for s in string_variable]
        clean_string_list = []
        invalid_strings_message = ""
        for string in string_variable:
            if cfg_key == "x_standardisation_joint":
                if string + "x" in data.columns:
                    clean_string_list.append(string)
                else:
                    invalid_strings_message += "\n" + string
            else:  # for y_standardisation_joint (& all other cases)!!
                if string + "y" in data.columns:
                    clean_string_list.append(string)
                else:
                    invalid_strings_message += "\n" + string
        if invalid_strings_message:
            # print and save warning
            strings_warning = (
                "\n***********\n! WARNING !\n***********\n"
                + "Unable to find all "
                + cfg_key
                + " joints/key points you entered!"
                + "\n\nInvalid and thus removed were:"
                + invalid_strings_message
            )
            print(strings_warning)
            write_issues_to_textfile(strings_warning, info)
            # clean the string
            string_variable = clean_string_list
            # print & save info about remaining ("clean") strings
            clean_strings_message = "\nRemaining joints / key points are:"
            for string in clean_string_list:
                clean_strings_message += "\n" + string
            clean_strings_message += (
                "\n\nNote that capitalisation matters."
                + "\nIf you are running a batch analysis, we'll use this updated cfg "
                + "throughout\nCheck out the config.json file for the full cfg used."
            )
            print(clean_strings_message)
            write_issues_to_textfile(clean_strings_message, info)

    # things are more involved for angle dicts
    elif type(string_variable) is dict:
        # 1) test if the lists of all keys are equally long, if not throw out last idxs
        key_lengths = [len(string_variable[key]) for key in string_variable.keys()]
        if not all(key_length == key_lengths[0] for key_length in key_lengths):
            min_length = min(key_lengths)
            for key in string_variable:  # remove invalid idxs
                string_variable[key] = string_variable[key][:min_length]
            key_length_mismatch_message = (  # inform user
                "\n***********\n! WARNING !\n***********\n"
                + "\nLength-mismatch in angle configuration!"
                + "\nCheck angles' name/upper/lower-joint entries."
                + "\nOnly processing first "
                + str(min_length)
                + " entries."
            )
            print(key_length_mismatch_message)
            write_issues_to_textfile(key_length_mismatch_message, info)
        # 2) check if all strings in the angle dict are valid columns
        invalid_angletrio_message = ""
        invalid_idxs = []  # these idxs hold across the 3 keys of our angles-dict
        for key in string_variable:
            # important to do this first (else string + "y" is invalid)
            string_variable[key] = [
                s if s.endswith(" ") else s + " " for s in string_variable[key]
            ]
            # remove any empty or invalid (i.e. not in data) strings from our angletrio
            this_keys_missing_strings = ""
            for idx, string in enumerate(string_variable[key]):
                if string + "y" not in data.columns:  # checks for empty strings too
                    invalid_idxs.append(idx)
                    if not this_keys_missing_strings:  # first occurance
                        this_keys_missing_strings += "\nAngle's " + key + " key: "
                    this_keys_missing_strings += string + "(#" + str(idx + 1) + ") / "
            # string concat outside of idx-forloop above please
            invalid_angletrio_message += this_keys_missing_strings
        # if we have to remove idxs from all keys of our angles dict
        if invalid_angletrio_message:
            # print and save
            angles_warning = (
                "\n***********\n! WARNING !\n***********\n"
                + "Unable to find all "
                + cfg_key
                + " joints/key points you entered!"
                + "\n\nInvalid and thus removed were:"
                + invalid_angletrio_message
            )
            print(angles_warning)
            write_issues_to_textfile(angles_warning, info)
            # clean the dict
            invalid_idxs = list(set(invalid_idxs))  # remove duplicates
            for key in string_variable:
                string_variable[key] = [
                    string
                    for i, string in enumerate(string_variable[key])
                    if i not in invalid_idxs
                ]
            # print & save info about the remaining ("clean") angles dict
            clean_angles_message = "\nRemaining angle configuration is:"
            for i in range(len(string_variable["name"])):
                clean_angles_message += "\n\nAngle #" + str(i + 1)
                clean_angles_message += "\nName: " + string_variable["name"][i]
                clean_angles_message += (
                    "\nLower Joint: " + string_variable["lower_joint"][i]
                )
                clean_angles_message += (
                    "\nUpper Joint: " + string_variable["upper_joint"][i]
                )
            clean_angles_message += (
                "\n\nNote that capitalisation matters."
                + "\nIf you are running a batch analysis, we'll use this updated cfg "
                + "throughout\nCheck out the config.json file for the full cfg used."
            )
            print(clean_angles_message)
            write_issues_to_textfile(clean_angles_message, info)

    return string_variable


def flip_mouse_body(data, info):
    """If the mouse ran through the video frame from right to left simulate
    that it ran from left to right. For this just subtract all x-values of
    all x-columns from their respective maxima.
    ==> This preserves time-information important for SC extraction via table
    ==> This preserves y-information too
    ==> All analyses & plots are therefore comparable to mice that did really
        run from left to right
    """

    # 0) Tell the user that we are flipping their mouse
    message = (
        "\nDetermined gait direction of right => left - simulating it to be left => "
        + "right"
    )
    print(message)
    write_issues_to_textfile(message, info)

    # 1) Flip all rows in x columns only and subtract max from all vals
    flipped_data = data.copy()
    x_cols = [col for col in flipped_data.columns if col.endswith(" x")]
    global_x_max = flipped_data[x_cols].max().max()
    for col in x_cols:
        flipped_data[col] = global_x_max - flipped_data[col]
    return flipped_data



from autogaita.universal3D.universal3D_1_preparation import some_prep
from autogaita.universal3D.universal3D_2_sc_extraction import (
    check_different_angle_joint_coords,
)
import os
import numpy as np
import pytest


# %%................................  fixtures  ........................................
@pytest.fixture
def extract_info(tmp_path):
    info = {}
    info["name"] = "TestSubject"
    info["results_dir"] = os.path.join(tmp_path, info["name"])
    return info


@pytest.fixture
def extract_folderinfo():
    folderinfo = {}
    folderinfo["root_dir"] = "tests/test_data/universal3D_data/test_data"
    folderinfo["sctable_filename"] = "SC Latency Table.xlsx"
    folderinfo["postname_string"] = ""
    return folderinfo


@pytest.fixture
def extract_cfg():
    # note space in end of "Midfoot, left " must be here bc. we don't run our fix and
    # check cfg function of 1_prep_
    cfg = {}
    cfg["sampling_rate"] = 100  # base cfg
    cfg["dont_show_plots"] = True
    cfg["y_acceleration"] = True
    cfg["angular_acceleration"] = True
    cfg["bin_num"] = 25
    cfg["plot_SE"] = True
    cfg["standardise_z_at_SC_level"] = True
    cfg["standardise_z_to_a_joint"] = True
    cfg["z_standardisation_joint"] = ["Midfoot, left "]
    cfg["plot_joint_number"] = 5
    cfg["color_palette"] = "Set2"
    cfg["legend_outside"] = True
    cfg["flip_gait_direction"] = True
    cfg["analyse_average_y"] = True
    cfg["standardise_y_coordinates"] = True
    cfg["y_standardisation_joint"] = ["Midfoot, left "]
    cfg["coordinate_standardisation_xls"] = ""
    cfg["joints"] = [
        "Midfoot",
        "Ankle",
        "Knee",
        "Hip",
    ]
    cfg["angles"] = {
        "name": ["Ankle", "Knee"],
        "lower_joint": ["Midfoot", "Ankle"],
        "upper_joint": ["Knee", "Hip"],
    }
    cfg["direction_joint"] = "Midfoot, left Y"
    return cfg


# ..................................  tests  .........................................


def test_clean_cycles_different_angle_joint_coords(
    extract_info, extract_folderinfo, extract_cfg
):
    # start conditions
    data, _ = some_prep(extract_info, extract_folderinfo, extract_cfg)
    all_cycles = [[[111, 222], [333, 444]], [[555, 666], [777, 888]]]
    # test one: leads to both lists still being non-empty, one SC removed
    data.loc[123, "Ankle, left Y"] = 5.5
    data.loc[123, "Ankle, left Z"] = 2.1
    data.loc[123, "Midfoot, left Y"] = 5.5
    data.loc[123, "Midfoot, left Z"] = 2.1
    all_cycles = check_different_angle_joint_coords(
        all_cycles, data, extract_info, extract_cfg
    )
    assert all_cycles == [[[333, 444]], [[555, 666], [777, 888]]]
    with open(os.path.join(extract_info["results_dir"], "Issues.txt")) as f:
        content = f.read()
    assert "Run #1 - SC #1" in content
    assert "equal LEFT" in content
    assert "Lower joint: Midfoot" in content
    # test two: leads to one empty list
    data.loc[345, "Hip, right Y"] = 10.1
    data.loc[345, "Hip, right Z"] = 0.333
    data.loc[345, "Ankle, right Y"] = 10.1
    data.loc[345, "Ankle, right Z"] = 0.333
    all_cycles = check_different_angle_joint_coords(
        all_cycles, data, extract_info, extract_cfg
    )
    assert all_cycles == [[], [[555, 666], [777, 888]]]
    with open(os.path.join(extract_info["results_dir"], "Issues.txt")) as f:
        content = f.read()
    assert "RIGHT" in content
    # test three: leads to all_cycles being None
    data.loc[[567, 789], "Hip, right Y"] = 22.2
    data.loc[[567, 789], "Hip, right Z"] = 0.11
    data.loc[[567, 789], "Knee, right Y"] = 22.2
    data.loc[[567, 789], "Knee, right Z"] = 0.11
    assert (
        check_different_angle_joint_coords(all_cycles, data, extract_info, extract_cfg)
        is None
    )
